"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)
    :codeauthor: Alexandru Bleotu (alexandru.bleotu@morganstanley.com)

    tests.unit.pillar_test
    ~~~~~~~~~~~~~~~~~~~~~~
"""


import logging
import os
import shutil
import tempfile
import textwrap

import salt.config
import salt.exceptions
import salt.fileclient
import salt.utils.stringutils
from salt.utils.files import fopen
from tests.support.helpers import with_tempdir
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class MockFileclient:
    def __init__(self, cache_file=None, get_state=None, list_states=None):
        if cache_file is not None:
            self.cache_file = lambda *x, **y: cache_file
        if get_state is not None:
            self.get_state = lambda sls, env: get_state[sls]
        if list_states is not None:
            self.list_states = lambda *x, **y: list_states

    # pylint: disable=unused-argument,no-method-argument,method-hidden
    def cache_file(*args, **kwargs):
        raise NotImplementedError()

    def get_state(*args, **kwargs):
        raise NotImplementedError()

    def list_states(*args, **kwargs):
        raise NotImplementedError()

    # pylint: enable=unused-argument,no-method-argument,method-hidden


class PillarTestCase(TestCase):
    def tearDown(self):
        for attrname in (
            "generic_file",
            "generic_minion_file",
            "ssh_file",
            "ssh_minion_file",
            "top_file",
        ):
            try:
                delattr(self, attrname)
            except AttributeError:
                continue

    def test_pillarenv_from_saltenv(self):
        with patch("salt.pillar.compile_template") as compile_template:
            opts = {
                "optimization_order": [0, 1, 2],
                "renderer": "json",
                "renderer_blacklist": [],
                "renderer_whitelist": [],
                "state_top": "",
                "pillar_roots": {"dev": [], "base": []},
                "file_roots": {"dev": [], "base": []},
                "extension_modules": "",
                "pillarenv_from_saltenv": True,
            }
            grains = {
                "os": "Ubuntu",
            }
            pillar = salt.pillar.Pillar(opts, grains, "mocked-minion", "dev")
            self.assertEqual(pillar.opts["saltenv"], "dev")
            self.assertEqual(pillar.opts["pillarenv"], "dev")

    def test_ext_pillar_no_extra_minion_data_val_dict(self):
        opts = {
            "optimization_order": [0, 1, 2],
            "renderer": "json",
            "renderer_blacklist": [],
            "renderer_whitelist": [],
            "state_top": "",
            "pillar_roots": {"dev": [], "base": []},
            "file_roots": {"dev": [], "base": []},
            "extension_modules": "",
            "pillarenv_from_saltenv": True,
        }
        mock_ext_pillar_func = MagicMock()
        with patch(
            "salt.loader.pillars",
            MagicMock(return_value={"fake_ext_pillar": mock_ext_pillar_func}),
        ):
            pillar = salt.pillar.Pillar(opts, {}, "mocked-minion", "dev")
        # ext pillar function doesn't have the extra_minion_data arg
        with patch(
            "salt.utils.args.get_function_argspec",
            MagicMock(return_value=MagicMock(args=[])),
        ):
            pillar._external_pillar_data(
                "fake_pillar", {"arg": "foo"}, "fake_ext_pillar"
            )
        mock_ext_pillar_func.assert_called_once_with(
            "mocked-minion", "fake_pillar", arg="foo"
        )
        # ext pillar function has the extra_minion_data arg
        mock_ext_pillar_func.reset_mock()
        with patch(
            "salt.utils.args.get_function_argspec",
            MagicMock(return_value=MagicMock(args=["extra_minion_data"])),
        ):
            pillar._external_pillar_data(
                "fake_pillar", {"arg": "foo"}, "fake_ext_pillar"
            )
        mock_ext_pillar_func.assert_called_once_with(
            "mocked-minion", "fake_pillar", arg="foo"
        )

    def test_ext_pillar_no_extra_minion_data_val_list(self):
        opts = {
            "optimization_order": [0, 1, 2],
            "renderer": "json",
            "renderer_blacklist": [],
            "renderer_whitelist": [],
            "state_top": "",
            "pillar_roots": {"dev": [], "base": []},
            "file_roots": {"dev": [], "base": []},
            "extension_modules": "",
            "pillarenv_from_saltenv": True,
        }
        mock_ext_pillar_func = MagicMock()
        with patch(
            "salt.loader.pillars",
            MagicMock(return_value={"fake_ext_pillar": mock_ext_pillar_func}),
        ):
            pillar = salt.pillar.Pillar(opts, {}, "mocked-minion", "dev")
        # ext pillar function doesn't have the extra_minion_data arg
        with patch(
            "salt.utils.args.get_function_argspec",
            MagicMock(return_value=MagicMock(args=[])),
        ):
            pillar._external_pillar_data("fake_pillar", ["foo"], "fake_ext_pillar")
        mock_ext_pillar_func.assert_called_once_with(
            "mocked-minion", "fake_pillar", "foo"
        )
        # ext pillar function has the extra_minion_data arg
        mock_ext_pillar_func.reset_mock()
        with patch(
            "salt.utils.args.get_function_argspec",
            MagicMock(return_value=MagicMock(args=["extra_minion_data"])),
        ):
            pillar._external_pillar_data("fake_pillar", ["foo"], "fake_ext_pillar")
        mock_ext_pillar_func.assert_called_once_with(
            "mocked-minion", "fake_pillar", "foo"
        )

    def test_ext_pillar_no_extra_minion_data_val_elem(self):
        opts = {
            "optimization_order": [0, 1, 2],
            "renderer": "json",
            "renderer_blacklist": [],
            "renderer_whitelist": [],
            "state_top": "",
            "pillar_roots": {"dev": [], "base": []},
            "file_roots": {"dev": [], "base": []},
            "extension_modules": "",
            "pillarenv_from_saltenv": True,
        }
        mock_ext_pillar_func = MagicMock()
        with patch(
            "salt.loader.pillars",
            MagicMock(return_value={"fake_ext_pillar": mock_ext_pillar_func}),
        ):
            pillar = salt.pillar.Pillar(opts, {}, "mocked-minion", "dev")
        # ext pillar function doesn't have the extra_minion_data arg
        with patch(
            "salt.utils.args.get_function_argspec",
            MagicMock(return_value=MagicMock(args=[])),
        ):
            pillar._external_pillar_data("fake_pillar", "fake_val", "fake_ext_pillar")
        mock_ext_pillar_func.assert_called_once_with(
            "mocked-minion", "fake_pillar", "fake_val"
        )
        # ext pillar function has the extra_minion_data arg
        mock_ext_pillar_func.reset_mock()
        with patch(
            "salt.utils.args.get_function_argspec",
            MagicMock(return_value=MagicMock(args=["extra_minion_data"])),
        ):
            pillar._external_pillar_data("fake_pillar", "fake_val", "fake_ext_pillar")
        mock_ext_pillar_func.assert_called_once_with(
            "mocked-minion", "fake_pillar", "fake_val"
        )

    def test_ext_pillar_with_extra_minion_data_val_dict(self):
        opts = {
            "optimization_order": [0, 1, 2],
            "renderer": "json",
            "renderer_blacklist": [],
            "renderer_whitelist": [],
            "state_top": "",
            "pillar_roots": {"dev": [], "base": []},
            "file_roots": {"dev": [], "base": []},
            "extension_modules": "",
            "pillarenv_from_saltenv": True,
        }
        mock_ext_pillar_func = MagicMock()
        with patch(
            "salt.loader.pillars",
            MagicMock(return_value={"fake_ext_pillar": mock_ext_pillar_func}),
        ):
            pillar = salt.pillar.Pillar(
                opts, {}, "mocked-minion", "dev", extra_minion_data={"fake_key": "foo"}
            )
        # ext pillar function doesn't have the extra_minion_data arg
        with patch(
            "salt.utils.args.get_function_argspec",
            MagicMock(return_value=MagicMock(args=[])),
        ):
            pillar._external_pillar_data(
                "fake_pillar", {"arg": "foo"}, "fake_ext_pillar"
            )
        mock_ext_pillar_func.assert_called_once_with(
            "mocked-minion", "fake_pillar", arg="foo"
        )
        # ext pillar function has the extra_minion_data arg
        mock_ext_pillar_func.reset_mock()
        with patch(
            "salt.utils.args.get_function_argspec",
            MagicMock(return_value=MagicMock(args=["extra_minion_data"])),
        ):
            pillar._external_pillar_data(
                "fake_pillar", {"arg": "foo"}, "fake_ext_pillar"
            )
        mock_ext_pillar_func.assert_called_once_with(
            "mocked-minion",
            "fake_pillar",
            arg="foo",
            extra_minion_data={"fake_key": "foo"},
        )

    def test_ext_pillar_with_extra_minion_data_val_list(self):
        opts = {
            "optimization_order": [0, 1, 2],
            "renderer": "json",
            "renderer_blacklist": [],
            "renderer_whitelist": [],
            "state_top": "",
            "pillar_roots": {"dev": [], "base": []},
            "file_roots": {"dev": [], "base": []},
            "extension_modules": "",
            "pillarenv_from_saltenv": True,
        }
        mock_ext_pillar_func = MagicMock()
        with patch(
            "salt.loader.pillars",
            MagicMock(return_value={"fake_ext_pillar": mock_ext_pillar_func}),
        ):
            pillar = salt.pillar.Pillar(
                opts, {}, "mocked-minion", "dev", extra_minion_data={"fake_key": "foo"}
            )
        # ext pillar function doesn't have the extra_minion_data arg
        with patch(
            "salt.utils.args.get_function_argspec",
            MagicMock(return_value=MagicMock(args=[])),
        ):
            pillar._external_pillar_data("fake_pillar", ["bar"], "fake_ext_pillar")
        mock_ext_pillar_func.assert_called_once_with(
            "mocked-minion", "fake_pillar", "bar"
        )
        # ext pillar function has the extra_minion_data arg
        mock_ext_pillar_func.reset_mock()
        with patch(
            "salt.utils.args.get_function_argspec",
            MagicMock(return_value=MagicMock(args=["extra_minion_data"])),
        ):
            pillar._external_pillar_data("fake_pillar", ["bar"], "fake_ext_pillar")
        mock_ext_pillar_func.assert_called_once_with(
            "mocked-minion", "fake_pillar", "bar", extra_minion_data={"fake_key": "foo"}
        )

    def test_ext_pillar_with_extra_minion_data_val_elem(self):
        opts = {
            "optimization_order": [0, 1, 2],
            "renderer": "json",
            "renderer_blacklist": [],
            "renderer_whitelist": [],
            "state_top": "",
            "pillar_roots": {"dev": [], "base": []},
            "file_roots": {"dev": [], "base": []},
            "extension_modules": "",
            "pillarenv_from_saltenv": True,
        }
        mock_ext_pillar_func = MagicMock()
        with patch(
            "salt.loader.pillars",
            MagicMock(return_value={"fake_ext_pillar": mock_ext_pillar_func}),
        ):
            pillar = salt.pillar.Pillar(
                opts, {}, "mocked-minion", "dev", extra_minion_data={"fake_key": "foo"}
            )
        # ext pillar function doesn't have the extra_minion_data arg
        with patch(
            "salt.utils.args.get_function_argspec",
            MagicMock(return_value=MagicMock(args=[])),
        ):
            pillar._external_pillar_data("fake_pillar", "bar", "fake_ext_pillar")
        mock_ext_pillar_func.assert_called_once_with(
            "mocked-minion", "fake_pillar", "bar"
        )
        # ext pillar function has the extra_minion_data arg
        mock_ext_pillar_func.reset_mock()
        with patch(
            "salt.utils.args.get_function_argspec",
            MagicMock(return_value=MagicMock(args=["extra_minion_data"])),
        ):
            pillar._external_pillar_data("fake_pillar", "bar", "fake_ext_pillar")
        mock_ext_pillar_func.assert_called_once_with(
            "mocked-minion", "fake_pillar", "bar", extra_minion_data={"fake_key": "foo"}
        )

    def test_ext_pillar_first(self):
        """
        test when using ext_pillar and ext_pillar_first
        """
        opts = {
            "optimization_order": [0, 1, 2],
            "renderer": "yaml",
            "renderer_blacklist": [],
            "renderer_whitelist": [],
            "state_top": "",
            "pillar_roots": [],
            "extension_modules": "",
            "saltenv": "base",
            "file_roots": [],
            "ext_pillar_first": True,
        }
        grains = {
            "os": "Ubuntu",
            "os_family": "Debian",
            "oscodename": "raring",
            "osfullname": "Ubuntu",
            "osrelease": "13.04",
            "kernel": "Linux",
        }

        tempdir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        try:
            sls_files = self._setup_test_topfile_sls_pillar_match(
                tempdir,
            )
            fc_mock = MockFileclient(
                cache_file=sls_files["top"]["dest"],
                list_states=["top", "ssh", "ssh.minion", "generic", "generic.minion"],
                get_state=sls_files,
            )
            with patch.object(
                salt.fileclient, "get_file_client", MagicMock(return_value=fc_mock)
            ), patch(
                "salt.pillar.Pillar.ext_pillar",
                MagicMock(
                    return_value=(
                        {"id": "minion", "phase": "alpha", "role": "database"},
                        [],
                    )
                ),
            ):
                pillar = salt.pillar.Pillar(opts, grains, "mocked-minion", "base")
                self.assertEqual(pillar.compile_pillar()["generic"]["key1"], "value1")
        finally:
            shutil.rmtree(tempdir, ignore_errors=True)

    @patch("salt.fileclient.Client.list_states")
    def test_malformed_pillar_sls(self, mock_list_states):
        with patch("salt.pillar.compile_template") as compile_template:
            opts = {
                "optimization_order": [0, 1, 2],
                "renderer": "json",
                "renderer_blacklist": [],
                "renderer_whitelist": [],
                "state_top": "",
                "pillar_roots": [],
                "file_roots": [],
                "extension_modules": "",
            }
            grains = {
                "os": "Ubuntu",
                "os_family": "Debian",
                "oscodename": "raring",
                "osfullname": "Ubuntu",
                "osrelease": "13.04",
                "kernel": "Linux",
            }

            mock_list_states.return_value = ["foo", "blah"]
            pillar = salt.pillar.Pillar(opts, grains, "mocked-minion", "base")
            # Mock getting the proper template files
            pillar.client.get_state = MagicMock(
                return_value={
                    "dest": "/path/to/pillar/files/foo.sls",
                    "source": "salt://foo.sls",
                }
            )

            # Template compilation returned a string
            compile_template.return_value = "BAHHH"
            self.assertEqual(
                pillar.render_pillar({"base": ["foo.sls"]}),
                ({}, ["SLS 'foo.sls' does not render to a dictionary"]),
            )

            # Template compilation returned a list
            compile_template.return_value = ["BAHHH"]
            self.assertEqual(
                pillar.render_pillar({"base": ["foo.sls"]}),
                ({}, ["SLS 'foo.sls' does not render to a dictionary"]),
            )

            # Template compilation returned a dictionary, which is what's expected
            compile_template.return_value = {"foo": "bar"}
            self.assertEqual(
                pillar.render_pillar({"base": ["foo.sls"]}), ({"foo": "bar"}, [])
            )

            # Test improper includes
            compile_template.side_effect = [
                {"foo": "bar", "include": "blah"},
                {"foo2": "bar2"},
            ]
            self.assertEqual(
                pillar.render_pillar({"base": ["foo.sls"]}),
                (
                    {"foo": "bar", "include": "blah"},
                    ["Include Declaration in SLS 'foo.sls' is not formed as a list"],
                ),
            )

            # Test includes as a list, which is what's expected
            compile_template.side_effect = [
                {"foo": "bar", "include": ["blah"]},
                {"foo2": "bar2"},
            ]
            self.assertEqual(
                pillar.render_pillar({"base": ["foo.sls"]}),
                ({"foo": "bar", "foo2": "bar2"}, []),
            )

            # Test includes as a list overriding data
            compile_template.side_effect = [
                {"foo": "bar", "include": ["blah"]},
                {"foo": "bar2"},
            ]
            self.assertEqual(
                pillar.render_pillar({"base": ["foo.sls"]}), ({"foo": "bar"}, [])
            )

            # Test includes using empty key directive
            compile_template.side_effect = [
                {"foo": "bar", "include": [{"blah": {"key": ""}}]},
                {"foo": "bar2"},
            ]
            self.assertEqual(
                pillar.render_pillar({"base": ["foo.sls"]}), ({"foo": "bar"}, [])
            )

            # Test includes using simple non-nested key
            compile_template.side_effect = [
                {"foo": "bar", "include": [{"blah": {"key": "nested"}}]},
                {"foo": "bar2"},
            ]
            self.assertEqual(
                pillar.render_pillar({"base": ["foo.sls"]}),
                ({"foo": "bar", "nested": {"foo": "bar2"}}, []),
            )

            # Test includes using nested key
            compile_template.side_effect = [
                {"foo": "bar", "include": [{"blah": {"key": "nested:level"}}]},
                {"foo": "bar2"},
            ]
            self.assertEqual(
                pillar.render_pillar({"base": ["foo.sls"]}),
                ({"foo": "bar", "nested": {"level": {"foo": "bar2"}}}, []),
            )

    def test_includes_override_sls(self):
        opts = {
            "optimization_order": [0, 1, 2],
            "renderer": "json",
            "renderer_blacklist": [],
            "renderer_whitelist": [],
            "state_top": "",
            "pillar_roots": {},
            "file_roots": {},
            "extension_modules": "",
        }
        grains = {
            "os": "Ubuntu",
            "os_family": "Debian",
            "oscodename": "raring",
            "osfullname": "Ubuntu",
            "osrelease": "13.04",
            "kernel": "Linux",
        }
        with patch("salt.pillar.compile_template") as compile_template, patch.object(
            salt.pillar.Pillar,
            "_Pillar__gather_avail",
            MagicMock(return_value={"base": ["blah", "foo"]}),
        ):

            # Test with option set to True
            opts["pillar_includes_override_sls"] = True
            pillar = salt.pillar.Pillar(opts, grains, "mocked-minion", "base")
            # Mock getting the proper template files
            pillar.client.get_state = MagicMock(
                return_value={
                    "dest": "/path/to/pillar/files/foo.sls",
                    "source": "salt://foo.sls",
                }
            )

            compile_template.side_effect = [
                {"foo": "bar", "include": ["blah"]},
                {"foo": "bar2"},
            ]
            self.assertEqual(
                pillar.render_pillar({"base": ["foo.sls"]}), ({"foo": "bar2"}, [])
            )

            # Test with option set to False
            opts["pillar_includes_override_sls"] = False
            pillar = salt.pillar.Pillar(opts, grains, "mocked-minion", "base")
            # Mock getting the proper template files
            pillar.client.get_state = MagicMock(
                return_value={
                    "dest": "/path/to/pillar/files/foo.sls",
                    "source": "salt://foo.sls",
                }
            )

            compile_template.side_effect = [
                {"foo": "bar", "include": ["blah"]},
                {"foo": "bar2"},
            ]
            self.assertEqual(
                pillar.render_pillar({"base": ["foo.sls"]}), ({"foo": "bar"}, [])
            )

    def test_topfile_order(self):
        opts = {
            "optimization_order": [0, 1, 2],
            "renderer": "yaml",
            "renderer_blacklist": [],
            "renderer_whitelist": [],
            "state_top": "",
            "pillar_roots": [],
            "extension_modules": "",
            "saltenv": "base",
            "file_roots": [],
        }
        grains = {
            "os": "Ubuntu",
            "os_family": "Debian",
            "oscodename": "raring",
            "osfullname": "Ubuntu",
            "osrelease": "13.04",
            "kernel": "Linux",
        }

        def _run_test(nodegroup_order, glob_order, expected):
            tempdir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
            try:
                sls_files = self._setup_test_topfile_sls(
                    tempdir, nodegroup_order, glob_order
                )
                fc_mock = MockFileclient(
                    cache_file=sls_files["top"]["dest"],
                    list_states=[
                        "top",
                        "ssh",
                        "ssh.minion",
                        "generic",
                        "generic.minion",
                    ],
                    get_state=sls_files,
                )
                with patch.object(
                    salt.fileclient, "get_file_client", MagicMock(return_value=fc_mock)
                ):
                    pillar = salt.pillar.Pillar(opts, grains, "mocked-minion", "base")
                    # Make sure that confirm_top.confirm_top returns True
                    pillar.matchers["confirm_top.confirm_top"] = lambda *x, **y: True
                    self.assertEqual(pillar.compile_pillar()["ssh"], expected)
            finally:
                shutil.rmtree(tempdir, ignore_errors=True)

        # test case where glob match happens second and therefore takes
        # precedence over nodegroup match.
        _run_test(nodegroup_order=1, glob_order=2, expected="bar")
        # test case where nodegroup match happens second and therefore takes
        # precedence over glob match.
        _run_test(nodegroup_order=2, glob_order=1, expected="foo")

    def _setup_test_topfile_sls_pillar_match(self, tempdir):
        # Write a simple topfile and two pillar state files
        top_file = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
        s = """
base:
  'phase:alpha':
    - match: pillar
    - generic
"""
        top_file.write(salt.utils.stringutils.to_bytes(s))
        top_file.flush()
        generic_file = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
        generic_file.write(
            b"""
generic:
    key1: value1
"""
        )
        generic_file.flush()
        return {
            "top": {"path": "", "dest": top_file.name},
            "generic": {"path": "", "dest": generic_file.name},
        }

    def _setup_test_topfile_sls(self, tempdir, nodegroup_order, glob_order):
        # Write a simple topfile and two pillar state files
        top_file = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
        s = """
base:
    group:
        - match: nodegroup
        - order: {nodegroup_order}
        - ssh
        - generic
    '*':
        - generic
    minion:
        - order: {glob_order}
        - ssh.minion
        - generic.minion
""".format(
            nodegroup_order=nodegroup_order, glob_order=glob_order
        )
        top_file.write(salt.utils.stringutils.to_bytes(s))
        top_file.flush()
        ssh_file = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
        ssh_file.write(
            b"""
ssh:
    foo
"""
        )
        ssh_file.flush()
        ssh_minion_file = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
        ssh_minion_file.write(
            b"""
ssh:
    bar
"""
        )
        ssh_minion_file.flush()
        generic_file = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
        generic_file.write(
            b"""
generic:
    key1:
      - value1
      - value2
    key2:
        sub_key1: []
"""
        )
        generic_file.flush()
        generic_minion_file = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
        generic_minion_file.write(
            b"""
generic:
    key1:
      - value3
    key2:
        sub_key2: []
"""
        )
        generic_minion_file.flush()

        return {
            "top": {"path": "", "dest": top_file.name},
            "ssh": {"path": "", "dest": ssh_file.name},
            "ssh.minion": {"path": "", "dest": ssh_minion_file.name},
            "generic": {"path": "", "dest": generic_file.name},
            "generic.minion": {"path": "", "dest": generic_minion_file.name},
        }

    @with_tempdir()
    def test_include(self, tempdir):
        opts = {
            "optimization_order": [0, 1, 2],
            "renderer": "yaml",
            "renderer_blacklist": [],
            "renderer_whitelist": [],
            "state_top": "",
            "pillar_roots": [],
            "extension_modules": "",
            "saltenv": "base",
            "file_roots": [],
        }
        grains = {
            "os": "Ubuntu",
            "os_family": "Debian",
            "oscodename": "raring",
            "osfullname": "Ubuntu",
            "osrelease": "13.04",
            "kernel": "Linux",
        }
        sls_files = self._setup_test_include_sls(tempdir)
        fc_mock = MockFileclient(
            cache_file=sls_files["top"]["dest"],
            get_state=sls_files,
            list_states=[
                "top",
                "test.init",
                "test.sub1",
                "test.sub2",
                "test.sub_wildcard_1",
                "test.sub_with_init_dot",
                "test.sub.with.slashes",
            ],
        )
        with patch.object(
            salt.fileclient, "get_file_client", MagicMock(return_value=fc_mock)
        ):
            pillar = salt.pillar.Pillar(opts, grains, "minion", "base")
            # Make sure that confirm_top.confirm_top returns True
            pillar.matchers["confirm_top.confirm_top"] = lambda *x, **y: True
            compiled_pillar = pillar.compile_pillar()
            self.assertEqual(compiled_pillar["foo_wildcard"], "bar_wildcard")
            self.assertEqual(compiled_pillar["foo1"], "bar1")
            self.assertEqual(compiled_pillar["foo2"], "bar2")
            self.assertEqual(compiled_pillar["sub_with_slashes"], "sub_slashes_worked")
            self.assertEqual(
                compiled_pillar["sub_init_dot"], "sub_with_init_dot_worked"
            )

    def _setup_test_include_sls(self, tempdir):
        top_file = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
        top_file.write(
            b"""
base:
    '*':
        - order: 1
        - test.sub2
    minion:
        - order: 2
        - test
"""
        )
        top_file.flush()
        init_sls = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
        init_sls.write(
            b"""
include:
   - test.sub1
   - test.sub_wildcard*
   - .test.sub_with_init_dot
   - test/sub/with/slashes
"""
        )
        init_sls.flush()
        sub1_sls = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
        sub1_sls.write(
            b"""
foo1:
  bar1
"""
        )
        sub1_sls.flush()
        sub2_sls = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
        sub2_sls.write(
            b"""
foo2:
  bar2
"""
        )
        sub2_sls.flush()

        sub_wildcard_1_sls = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
        sub_wildcard_1_sls.write(
            b"""
foo_wildcard:
   bar_wildcard
"""
        )
        sub_wildcard_1_sls.flush()

        sub_with_init_dot_sls = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
        sub_with_init_dot_sls.write(
            b"""
sub_init_dot:
  sub_with_init_dot_worked
"""
        )
        sub_with_init_dot_sls.flush()

        sub_with_slashes_sls = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
        sub_with_slashes_sls.write(
            b"""
sub_with_slashes:
  sub_slashes_worked
"""
        )
        sub_with_slashes_sls.flush()

        return {
            "top": {"path": "", "dest": top_file.name},
            "test": {"path": "", "dest": init_sls.name},
            "test.sub1": {"path": "", "dest": sub1_sls.name},
            "test.sub2": {"path": "", "dest": sub2_sls.name},
            "test.sub_wildcard_1": {"path": "", "dest": sub_wildcard_1_sls.name},
            "test.sub_with_init_dot": {"path": "", "dest": sub_with_init_dot_sls.name},
            "test.sub.with.slashes": {"path": "", "dest": sub_with_slashes_sls.name},
        }

    @with_tempdir()
    def test_relative_include(self, tempdir):
        join = os.path.join
        with fopen(join(tempdir, "top.sls"), "w") as f:
            print(
                textwrap.dedent(
                    """
                    base:
                      '*':
                        - includer
                        - simple_includer
                        - includes.with.more.depth
                """
                ),
                file=f,
            )
        includer_dir = join(tempdir, "includer")
        os.makedirs(includer_dir)
        with fopen(join(includer_dir, "init.sls"), "w") as f:
            print(
                textwrap.dedent(
                    """
                    include:
                      - .this
                      - includer.that
                """
                ),
                file=f,
            )
        with fopen(join(includer_dir, "this.sls"), "w") as f:
            print(
                textwrap.dedent(
                    """
                    this:
                        is all good
                """
                ),
                file=f,
            )
        with fopen(join(includer_dir, "that.sls"), "w") as f:
            print(
                textwrap.dedent(
                    """
                    that:
                        is also all good
                """
                ),
                file=f,
            )

        with fopen(join(tempdir, "simple_includer.sls"), "w") as simpleincluder:
            print(
                textwrap.dedent(
                    """
                    include:
                      - .simple
                      - super_simple
                """
                ),
                file=simpleincluder,
            )
        with fopen(join(tempdir, "simple.sls"), "w") as f:
            print(
                textwrap.dedent(
                    """
                    simple:
                      simon
                """
                ),
                file=f,
            )
        with fopen(join(tempdir, "super_simple.sls"), "w") as f:
            print(
                textwrap.dedent(
                    """
                    super simple:
                      a caveman
                """
                ),
                file=f,
            )

        depth_dir = join(tempdir, "includes", "with", "more")
        os.makedirs(depth_dir)
        with fopen(join(depth_dir, "depth.sls"), "w") as f:
            print(
                textwrap.dedent(
                    """
                    include:
                      - .ramble
                      - includes.with.more.doors

                    mordor:
                        has dark depths
                """
                ),
                file=f,
            )

        with fopen(join(depth_dir, "ramble.sls"), "w") as f:
            print(
                textwrap.dedent(
                    """
                    found:
                        my precious
                """
                ),
                file=f,
            )

        with fopen(join(depth_dir, "doors.sls"), "w") as f:
            print(
                textwrap.dedent(
                    """
                    mojo:
                        bad risin'
                """
                ),
                file=f,
            )
        opts = {
            "optimization_order": [0, 1, 2],
            "renderer": "yaml",
            "renderer_blacklist": [],
            "renderer_whitelist": [],
            "state_top": "top.sls",
            "pillar_roots": {"base": [tempdir]},
            "extension_modules": "",
            "saltenv": "base",
            "file_roots": [],
            "file_ignore_regex": None,
            "file_ignore_glob": None,
        }
        grains = {
            "os": "Ubuntu",
            "os_family": "Debian",
            "oscodename": "raring",
            "osfullname": "Ubuntu",
            "osrelease": "13.04",
            "kernel": "Linux",
        }
        pillar = salt.pillar.Pillar(opts, grains, "minion", "base")
        # Make sure that confirm_top.confirm_top returns True
        pillar.matchers["confirm_top.confirm_top"] = lambda *x, **y: True

        # Act
        compiled_pillar = pillar.compile_pillar()

        # Assert
        self.assertEqual(compiled_pillar["this"], "is all good")
        self.assertEqual(compiled_pillar["that"], "is also all good")
        self.assertEqual(compiled_pillar["simple"], "simon")
        self.assertEqual(compiled_pillar["super simple"], "a caveman")
        self.assertEqual(compiled_pillar["mordor"], "has dark depths")
        self.assertEqual(compiled_pillar["found"], "my precious")
        self.assertEqual(compiled_pillar["mojo"], "bad risin'")

    @with_tempdir()
    def test_missing_include(self, tempdir):
        opts = {
            "optimization_order": [0, 1, 2],
            "renderer": "yaml",
            "renderer_blacklist": [],
            "renderer_whitelist": [],
            "state_top": "top.sls",
            "pillar_roots": {"base": [tempdir]},
            "extension_modules": "",
            "saltenv": "base",
            "file_roots": [],
            "file_ignore_regex": None,
            "file_ignore_glob": None,
        }
        grains = {
            "os": "Ubuntu",
            "os_family": "Debian",
            "oscodename": "raring",
            "osfullname": "Ubuntu",
            "osrelease": "13.04",
            "kernel": "Linux",
        }

        join = os.path.join
        with fopen(join(tempdir, "top.sls"), "w") as f:
            print(
                textwrap.dedent(
                    """
                    base:
                      '*':
                        - simple_include
                """
                ),
                file=f,
            )
        include_dir = join(tempdir, "simple_include")
        os.makedirs(include_dir)
        with fopen(join(include_dir, "init.sls"), "w") as f:
            print(
                textwrap.dedent(
                    """
                    include:
                      - simple_include.missing_include
                    simple_include: is ok
                    """
                ),
                file=f,
            )

        pillar = salt.pillar.Pillar(opts, grains, "minion", "base")
        # Make sure that confirm_top.confirm_top returns True
        pillar.matchers["confirm_top.confirm_top"] = lambda *x, **y: True

        # Act
        compiled_pillar = pillar.compile_pillar()

        # Assert
        self.assertEqual(compiled_pillar["simple_include"], "is ok")
        self.assertTrue("_errors" in compiled_pillar)
        self.assertTrue(
            "simple_include.missing_include" in compiled_pillar["_errors"][0]
        )


@patch("salt.channel.client.ReqChannel.factory", MagicMock())
class RemotePillarTestCase(TestCase):
    """
    Tests for instantiating a RemotePillar in salt.pillar
    """

    def setUp(self):
        self.grains = {}
        self.tmp_pki = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.opts = {
            "pki_dir": self.tmp_pki,
            "id": "minion",
            "master_uri": "tcp://127.0.0.1:4505",
            "__role": "minion",
            "keysize": 2048,
            "renderer": "json",
            "path_to_add": "fake_data",
            "path_to_add2": {"fake_data2": ["fake_data3", "fake_data4"]},
            "pass_to_ext_pillars": ["path_to_add", "path_to_add2"],
        }

    def tearDown(self):
        for attr in ("grains", "tmp_pki", "opts"):
            try:
                delattr(self, attr)
            except AttributeError:
                continue

    def test_get_opts_in_pillar_override_call(self):
        mock_get_extra_minion_data = MagicMock(return_value={})
        with patch(
            "salt.pillar.RemotePillarMixin.get_ext_pillar_extra_minion_data",
            mock_get_extra_minion_data,
        ):

            salt.pillar.RemotePillar(self.opts, self.grains, "mocked-minion", "dev")
        call_opts = dict(self.opts, saltenv="dev")
        mock_get_extra_minion_data.assert_called_once_with(call_opts)

    def test_multiple_keys_in_opts_added_to_pillar(self):
        opts = {
            "pki_dir": self.tmp_pki,
            "id": "minion",
            "master_uri": "tcp://127.0.0.1:4505",
            "__role": "minion",
            "keysize": 2048,
            "renderer": "json",
            "path_to_add": "fake_data",
            "path_to_add2": {"fake_data2": ["fake_data3", "fake_data4"]},
            "pass_to_ext_pillars": ["path_to_add", "path_to_add2"],
        }
        pillar = salt.pillar.RemotePillar(
            self.opts, self.grains, "mocked-minion", "dev"
        )
        self.assertEqual(
            pillar.extra_minion_data,
            {
                "path_to_add": "fake_data",
                "path_to_add2": {"fake_data2": ["fake_data3", "fake_data4"]},
            },
        )

    def test_subkey_in_opts_added_to_pillar(self):
        opts = dict(
            self.opts,
            path_to_add2={
                "fake_data5": "fake_data6",
                "fake_data2": ["fake_data3", "fake_data4"],
            },
            pass_to_ext_pillars=["path_to_add2:fake_data5"],
        )
        pillar = salt.pillar.RemotePillar(opts, self.grains, "mocked-minion", "dev")
        self.assertEqual(
            pillar.extra_minion_data, {"path_to_add2": {"fake_data5": "fake_data6"}}
        )

    def test_non_existent_leaf_opt_in_add_to_pillar(self):
        pillar = salt.pillar.RemotePillar(
            self.opts, self.grains, "mocked-minion", "dev"
        )
        self.assertEqual(pillar.pillar_override, {})

    def test_non_existent_intermediate_opt_in_add_to_pillar(self):
        pillar = salt.pillar.RemotePillar(
            self.opts, self.grains, "mocked-minion", "dev"
        )
        self.assertEqual(pillar.pillar_override, {})

    def test_malformed_add_to_pillar(self):
        opts = dict(self.opts, pass_to_ext_pillars=MagicMock())
        with self.assertRaises(salt.exceptions.SaltClientError) as excinfo:
            salt.pillar.RemotePillar(opts, self.grains, "mocked-minion", "dev")
        self.assertEqual(
            excinfo.exception.strerror, "'pass_to_ext_pillars' config is malformed."
        )

    def test_pillar_send_extra_minion_data_from_config(self):
        opts = {
            "pki_dir": self.tmp_pki,
            "id": "minion",
            "master_uri": "tcp://127.0.0.1:4505",
            "__role": "minion",
            "keysize": 2048,
            "renderer": "json",
            "pillarenv": "fake_pillar_env",
            "path_to_add": "fake_data",
            "path_to_add2": {
                "fake_data5": "fake_data6",
                "fake_data2": ["fake_data3", "fake_data4"],
            },
            "pass_to_ext_pillars": ["path_to_add"],
        }
        mock_channel = MagicMock(
            crypted_transfer_decode_dictentry=MagicMock(return_value={})
        )
        with patch(
            "salt.channel.client.ReqChannel.factory",
            MagicMock(return_value=mock_channel),
        ):
            pillar = salt.pillar.RemotePillar(
                opts, self.grains, "mocked_minion", "fake_env"
            )

        ret = pillar.compile_pillar()
        self.assertEqual(pillar.channel, mock_channel)
        mock_channel.crypted_transfer_decode_dictentry.assert_called_once_with(
            {
                "cmd": "_pillar",
                "ver": "2",
                "id": "mocked_minion",
                "grains": {},
                "saltenv": "fake_env",
                "pillarenv": "fake_pillar_env",
                "pillar_override": {},
                "extra_minion_data": {"path_to_add": "fake_data"},
            },
            dictkey="pillar",
        )

    def test_pillar_file_client_master_remote(self):
        """
        Test condition where local file_client and use_master_when_local option
        returns a remote file client.
        """
        mocked_minion = MagicMock()
        opts = {
            "pki_dir": self.tmp_pki,
            "id": "minion",
            "master_uri": "tcp://127.0.0.1:4505",
            "__role": "minion",
            "keysize": 2048,
            "file_client": "local",
            "use_master_when_local": True,
            "pillar_cache": None,
        }
        pillar = salt.pillar.get_pillar(opts, self.grains, mocked_minion)
        self.assertEqual(type(pillar), salt.pillar.RemotePillar)
        self.assertNotEqual(type(pillar), salt.pillar.PillarCache)


@patch("salt.channel.client.AsyncReqChannel.factory", MagicMock())
class AsyncRemotePillarTestCase(TestCase):
    """
    Tests for instantiating a AsyncRemotePillar in salt.pillar
    """

    def setUp(self):
        self.grains = {}
        self.tmp_pki = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)

    def tearDown(self):
        for attr in ("grains", "tmp_pki"):
            try:
                delattr(self, attr)
            except AttributeError:
                continue

    def test_get_opts_in_pillar_override_call(self):
        mock_get_extra_minion_data = MagicMock(return_value={})
        with patch(
            "salt.pillar.RemotePillarMixin.get_ext_pillar_extra_minion_data",
            mock_get_extra_minion_data,
        ):

            salt.pillar.RemotePillar({}, self.grains, "mocked-minion", "dev")
        mock_get_extra_minion_data.assert_called_once_with({"saltenv": "dev"})

    def test_pillar_send_extra_minion_data_from_config(self):
        opts = {
            "pki_dir": self.tmp_pki,
            "id": "minion",
            "master_uri": "tcp://127.0.0.1:4505",
            "__role": "minion",
            "keysize": 2048,
            "renderer": "json",
            "pillarenv": "fake_pillar_env",
            "path_to_add": "fake_data",
            "path_to_add2": {
                "fake_data5": "fake_data6",
                "fake_data2": ["fake_data3", "fake_data4"],
            },
            "pass_to_ext_pillars": ["path_to_add"],
        }
        mock_channel = MagicMock(
            crypted_transfer_decode_dictentry=MagicMock(return_value={})
        )
        with patch(
            "salt.channel.client.AsyncReqChannel.factory",
            MagicMock(return_value=mock_channel),
        ):
            pillar = salt.pillar.RemotePillar(
                opts, self.grains, "mocked_minion", "fake_env"
            )

        ret = pillar.compile_pillar()
        mock_channel.crypted_transfer_decode_dictentry.assert_called_once_with(
            {
                "cmd": "_pillar",
                "ver": "2",
                "id": "mocked_minion",
                "grains": {},
                "saltenv": "fake_env",
                "pillarenv": "fake_pillar_env",
                "pillar_override": {},
                "extra_minion_data": {"path_to_add": "fake_data"},
            },
            dictkey="pillar",
        )


@patch("salt.channel.client.ReqChannel.factory", MagicMock())
class PillarCacheTestCase(TestCase):
    """
    Tests for instantiating a PillarCache in salt.pillar
    """

    def setUp(self):
        self.grains = {}

    @classmethod
    def setUpClass(cls):
        cls.mock_master_default_opts = salt.config.DEFAULT_MASTER_OPTS.copy()
        cls.mock_master_default_opts["cachedir"] = tempfile.mkdtemp(
            dir=RUNTIME_VARS.TMP
        )

    def tearDown(self):
        for attr in ("grains",):
            try:
                delattr(self, attr)
            except AttributeError:
                continue

    def test_compile_pillar_disk_cache(self):
        self.mock_master_default_opts.update(
            {"pillar_cache_backend": "disk", "pillar_cache_ttl": 3600}
        )

        pillar = salt.pillar.PillarCache(
            self.mock_master_default_opts,
            self.grains,
            "mocked_minion",
            "fake_env",
            pillarenv="base",
        )

        with patch("salt.utils.cache.CacheDisk._write", MagicMock()):
            with patch(
                "salt.pillar.PillarCache.fetch_pillar",
                side_effect=[{"foo": "bar"}, {"foo": "baz"}],
            ):
                # Run once for pillarenv base
                ret = pillar.compile_pillar()
                expected_cache = {"mocked_minion": {"base": {"foo": "bar"}}}
                self.assertEqual(pillar.cache._dict, expected_cache)

                # Run a second time for pillarenv base
                ret = pillar.compile_pillar()
                expected_cache = {"mocked_minion": {"base": {"foo": "bar"}}}
                self.assertEqual(pillar.cache._dict, expected_cache)

                # Change the pillarenv
                pillar.pillarenv = "dev"

                # Run once for pillarenv dev
                ret = pillar.compile_pillar()
                expected_cache = {
                    "mocked_minion": {"base": {"foo": "bar"}, "dev": {"foo": "baz"}}
                }
                self.assertEqual(pillar.cache._dict, expected_cache)

                # Run a second time for pillarenv dev
                ret = pillar.compile_pillar()
                expected_cache = {
                    "mocked_minion": {"base": {"foo": "bar"}, "dev": {"foo": "baz"}}
                }
                self.assertEqual(pillar.cache._dict, expected_cache)

    def test_compile_pillar_memory_cache(self):
        self.mock_master_default_opts.update(
            {"pillar_cache_backend": "memory", "pillar_cache_ttl": 3600}
        )

        pillar = salt.pillar.PillarCache(
            self.mock_master_default_opts,
            self.grains,
            "mocked_minion",
            "fake_env",
            pillarenv="base",
        )

        with patch(
            "salt.pillar.PillarCache.fetch_pillar",
            side_effect=[{"foo": "bar"}, {"foo": "baz"}],
        ):
            # Run once for pillarenv base
            ret = pillar.compile_pillar()
            expected_cache = {"base": {"foo": "bar"}}
            self.assertIn("mocked_minion", pillar.cache)
            self.assertEqual(pillar.cache["mocked_minion"], expected_cache)

            # Run a second time for pillarenv base
            ret = pillar.compile_pillar()
            expected_cache = {"base": {"foo": "bar"}}
            self.assertIn("mocked_minion", pillar.cache)
            self.assertEqual(pillar.cache["mocked_minion"], expected_cache)

            # Change the pillarenv
            pillar.pillarenv = "dev"

            # Run once for pillarenv dev
            ret = pillar.compile_pillar()
            expected_cache = {"base": {"foo": "bar"}, "dev": {"foo": "baz"}}
            self.assertIn("mocked_minion", pillar.cache)
            self.assertEqual(pillar.cache["mocked_minion"], expected_cache)

            # Run a second time for pillarenv dev
            ret = pillar.compile_pillar()
            expected_cache = {"base": {"foo": "bar"}, "dev": {"foo": "baz"}}
            self.assertIn("mocked_minion", pillar.cache)
            self.assertEqual(pillar.cache["mocked_minion"], expected_cache)



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""
import textwrap

import pytest

import salt.modules.mac_brew_pkg as mac_brew
import salt.utils.pkg
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, Mock, patch


@pytest.fixture
def TAPS_STRING():
    return "homebrew/dupes\nhomebrew/science\nhomebrew/x11"


@pytest.fixture
def TAPS_LIST():
    return ["homebrew/dupes", "homebrew/science", "homebrew/x11"]


@pytest.fixture
def HOMEBREW_BIN():
    return "/usr/local/bin/brew"


@pytest.fixture
def configure_loader_modules():
    return {mac_brew: {"__opts__": {"user": MagicMock(return_value="bar")}}}


def custom_call_brew(*cmd, failhard=True):
    result = dict()
    if cmd == ("info", "--json=v2", "--installed"):
        result = {
            "stdout": textwrap.dedent(
                """\
                {
                  "casks": [
                    {
                      "appcast": null,
                      "artifacts": [
                        [
                          "Day-3.0/Day-O.app"
                        ],
                        {
                          "signal": {},
                          "trash": "~/Library/Preferences/com.shauninman.Day-O.plist"
                        }
                      ],
                      "auto_updates": null,
                      "caveats": null,
                      "conflicts_with": null,
                      "container": null,
                      "depends_on": {},
                      "desc": null,
                      "homepage": "https://shauninman.com/archive/2020/04/08/day_o_mac_menu_bar_clock_for_catalina",
                      "installed": "3.0.1",
                      "name": [
                        "Day-O"
                      ],
                      "outdated": false,
                      "sha256": "4963f503c1e47bfa0f8bdbbbe5694d6a7242d298fb44ff68af80d42f1eaebaf9",
                      "token": "day-o",
                      "full_token": "day-o",
                      "tap": "homebrew/cask",
                      "url": "https://shauninman.com/assets/downloads/Day-3.0.zip",
                      "version": "3.0.1"
                    },
                    {
                      "appcast": null,
                      "artifacts": [
                        [
                          "iTerm.app"
                        ],
                        {
                          "signal": {},
                          "trash": [
                            "~/Library/Application Support/iTerm",
                            "~/Library/Application Support/iTerm2",
                            "~/Library/Application Support/com.apple.sharedfilelist/com.apple.LSSharedFileList.ApplicationRecentDocuments/com.googlecode.iterm2.sfl*",
                            "~/Library/Caches/com.googlecode.iterm2",
                            "~/Library/Preferences/com.googlecode.iterm2.plist",
                            "~/Library/Saved Application State/com.googlecode.iterm2.savedState"
                          ]
                        }
                      ],
                      "auto_updates": true,
                      "caveats": null,
                      "conflicts_with": {
                        "cask": [
                          "iterm2-beta"
                        ]
                      },
                      "container": null,
                      "depends_on": {
                        "macos": {
                          ">=": [
                            "10.12"
                          ]
                        }
                      },
                      "desc": "Terminal emulator as alternative to Apple's Terminal app",
                      "homepage": "https://www.iterm2.com/",
                      "installed": "3.4.3",
                      "name": [
                        "iTerm2"
                      ],
                      "outdated": false,
                      "sha256": "9ed73844838bddf797eadf37e5f7da3771308c3f74d38cd422c18eebaaa8f6b9",
                      "token": "iterm2",
                      "full_token": "custom/tap/iterm2",
                      "tap": "custom/tap",
                      "url": "https://iterm2.com/downloads/stable/iTerm2-3_4_3.zip",
                      "version": "3.4.3"
                    },
                    {
                      "token": "discord",
                      "full_token": "discord",
                      "tap": null,
                      "name": [
                        "Discord"
                      ],
                      "desc": "Voice and text chat software",
                      "homepage": "https://discord.com/",
                      "url": "https://dl.discordapp.net/apps/osx/0.0.268/Discord.dmg",
                      "appcast": null,
                      "version": "0.0.268",
                      "versions": {
                      },
                      "installed": "0.0.266",
                      "outdated": false,
                      "sha256": "dfe12315b717ed06ac24d3eaacb700618e96cbb449ed63d2afadcdb70ad09c55",
                      "artifacts": [
                        {
                          "app": [
                            "Discord.app"
                          ]
                        },
                        {
                          "zap": [
                            {
                              "trash": [
                                "~/Library/Application Support/discord",
                                "~/Library/Caches/com.hnc.Discord",
                                "~/Library/Caches/com.hnc.Discord.ShipIt",
                                "~/Library/Cookies/com.hnc.Discord.binarycookies",
                                "~/Library/Preferences/com.hnc.Discord.helper.plist",
                                "~/Library/Preferences/com.hnc.Discord.plist",
                                "~/Library/Saved Application State/com.hnc.Discord.savedState"
                              ]
                            }
                          ]
                        }
                      ],
                      "caveats": null,
                      "depends_on": {
                      },
                      "conflicts_with": null,
                      "container": null,
                      "auto_updates": true
                    }
                  ],
                  "formulae": [
                    {
                      "aliases": [],
                      "bottle": {
                        "stable": {
                          "cellar": ":any",
                          "files": {
                            "arm64_big_sur": {
                              "sha256": "674b3ae41c399f1e8e44c271b0e6909babff9fcd2e04a2127d25e2407ea4dd33",
                              "url": "https://homebrew.bintray.com/bottles/jq-1.6.arm64_big_sur.bottle.1.tar.gz"
                            },
                            "big_sur": {
                              "sha256": "bf0f8577632af7b878b6425476f5b1ab9c3bf66d65affb0c455048a173a0b6bf",
                              "url": "https://homebrew.bintray.com/bottles/jq-1.6.big_sur.bottle.1.tar.gz"
                            },
                            "catalina": {
                              "sha256": "820a3c85fcbb63088b160c7edf125d7e55fc2c5c1d51569304499c9cc4b89ce8",
                              "url": "https://homebrew.bintray.com/bottles/jq-1.6.catalina.bottle.1.tar.gz"
                            },
                            "high_sierra": {
                              "sha256": "dffcffa4ea13e8f0f2b45c5121e529077e135ae9a47254c32182231662ee9b72",
                              "url": "https://homebrew.bintray.com/bottles/jq-1.6.high_sierra.bottle.1.tar.gz"
                            },
                            "mojave": {
                              "sha256": "71f0e76c5b22e5088426c971d5e795fe67abee7af6c2c4ae0cf4c0eb98ed21ff",
                              "url": "https://homebrew.bintray.com/bottles/jq-1.6.mojave.bottle.1.tar.gz"
                            },
                            "sierra": {
                              "sha256": "bb4d19dc026c2d72c53eed78eaa0ab982e9fcad2cd2acc6d13e7a12ff658e877",
                              "url": "https://homebrew.bintray.com/bottles/jq-1.6.sierra.bottle.1.tar.gz"
                            }
                          },
                          "prefix": "/usr/local",
                          "rebuild": 1,
                          "root_url": "https://homebrew.bintray.com/bottles"
                        }
                      },
                      "bottle_disabled": false,
                      "build_dependencies": [],
                      "caveats": null,
                      "conflicts_with": [],
                      "dependencies": [
                        "oniguruma"
                      ],
                      "deprecated": false,
                      "deprecation_date": null,
                      "deprecation_reason": null,
                      "desc": "Lightweight and flexible command-line JSON processor",
                      "disable_date": null,
                      "disable_reason": null,
                      "disabled": false,
                      "full_name": "jq",
                      "homepage": "https://stedolan.github.io/jq/",
                      "installed": [
                        {
                          "built_as_bottle": true,
                          "installed_as_dependency": false,
                          "installed_on_request": true,
                          "poured_from_bottle": true,
                          "runtime_dependencies": [
                            {
                              "full_name": "oniguruma",
                              "version": "6.9.6"
                            }
                          ],
                          "used_options": [],
                          "version": "1.6"
                        }
                      ],
                      "keg_only": false,
                      "license": "MIT",
                      "linked_keg": "1.6",
                      "name": "jq",
                      "oldname": null,
                      "optional_dependencies": [],
                      "options": [],
                      "outdated": false,
                      "pinned": false,
                      "recommended_dependencies": [],
                      "requirements": [],
                      "revision": 0,
                      "urls": {
                        "stable": {
                          "revision": null,
                          "tag": null,
                          "url": "https://github.com/stedolan/jq/releases/download/jq-1.6/jq-1.6.tar.gz"
                        }
                      },
                      "uses_from_macos": [],
                      "version_scheme": 0,
                      "versioned_formulae": [],
                      "versions": {
                        "bottle": true,
                        "head": "HEAD",
                        "stable": "1.6"
                      }
                    },
                    {
                      "aliases": [],
                      "bottle": {
                        "stable": {
                          "cellar": ":any",
                          "files": {
                            "arm64_big_sur": {
                              "sha256": "c84206005787304416ed81094bd3a0cdd2ae8eb62649db5a3a44fa14b276d09f",
                              "url": "https://homebrew.bintray.com/bottles/xz-5.2.5.arm64_big_sur.bottle.tar.gz"
                            },
                            "big_sur": {
                              "sha256": "4fbd4a9e3eb49c27e83bd125b0e76d386c0e12ae1139d4dc9e31841fb8880a35",
                              "url": "https://homebrew.bintray.com/bottles/xz-5.2.5.big_sur.bottle.tar.gz"
                            },
                            "catalina": {
                              "sha256": "2dcc8e0121c934d1e34ffdb37fcd70f0f7b5c2f4755f2f7cbcf360e9e54cb43b",
                              "url": "https://homebrew.bintray.com/bottles/xz-5.2.5.catalina.bottle.tar.gz"
                            },
                            "high_sierra": {
                              "sha256": "1491b2b20c40c3cb0b990f520768d7e876e4ab4a7dc1da9994d0150da34ba5c6",
                              "url": "https://homebrew.bintray.com/bottles/xz-5.2.5.high_sierra.bottle.tar.gz"
                            },
                            "mojave": {
                              "sha256": "44483961b5d2b535b0ece1936c9d40b4bc7d9c7281646cca0fb476291ab9d4dc",
                              "url": "https://homebrew.bintray.com/bottles/xz-5.2.5.mojave.bottle.tar.gz"
                            }
                          },
                          "prefix": "/usr/local",
                          "rebuild": 0,
                          "root_url": "https://homebrew.bintray.com/bottles"
                        }
                      },
                      "bottle_disabled": false,
                      "build_dependencies": [],
                      "caveats": null,
                      "conflicts_with": [],
                      "dependencies": [],
                      "deprecated": false,
                      "deprecation_date": null,
                      "deprecation_reason": null,
                      "desc": "General-purpose data compression with high compression ratio",
                      "disable_date": null,
                      "disable_reason": null,
                      "disabled": false,
                      "full_name": "xz",
                      "homepage": "https://tukaani.org/xz/",
                      "installed": [
                        {
                          "built_as_bottle": true,
                          "installed_as_dependency": true,
                          "installed_on_request": false,
                          "poured_from_bottle": true,
                          "runtime_dependencies": [],
                          "used_options": [],
                          "version": "5.2.5"
                        }
                      ],
                      "keg_only": false,
                      "license": "GPL-2.0",
                      "linked_keg": "5.2.5",
                      "name": "xz",
                      "oldname": null,
                      "optional_dependencies": [],
                      "options": [],
                      "outdated": false,
                      "pinned": false,
                      "recommended_dependencies": [],
                      "requirements": [],
                      "revision": 0,
                      "urls": {
                        "stable": {
                          "revision": null,
                          "tag": null,
                          "url": "https://downloads.sourceforge.net/project/lzmautils/xz-5.2.5.tar.gz"
                        }
                      },
                      "uses_from_macos": [],
                      "version_scheme": 0,
                      "versioned_formulae": [],
                      "versions": {
                        "bottle": true,
                        "head": null,
                        "stable": "5.2.5"
                      }
                    }
                  ]
                }
                """
            ),
            "stderr": "",
            "retcode": 0,
        }

    return result


def custom_add_pkg(ret, name, newest_version):
    ret[name] = newest_version
    return ret


# '_list_taps' function tests: 1


def test_list_taps(TAPS_STRING, TAPS_LIST, HOMEBREW_BIN):
    """
    Tests the return of the list of taps
    """
    mock_taps = MagicMock(return_value={"stdout": TAPS_STRING, "retcode": 0})
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    with patch("salt.utils.path.which", MagicMock(return_value=HOMEBREW_BIN)):
        with patch.dict(
            mac_brew.__salt__,
            {"file.get_user": mock_user, "cmd.run_all": mock_taps, "cmd.run": mock_cmd},
        ):
            assert mac_brew._list_taps() == TAPS_LIST


# '_tap' function tests: 3


def test_tap_installed(TAPS_LIST):
    """
    Tests if tap argument is already installed or not
    """
    with patch(
        "salt.modules.mac_brew_pkg._list_taps", MagicMock(return_value=TAPS_LIST)
    ):
        assert mac_brew._tap("homebrew/science")


def test_tap_failure(HOMEBREW_BIN):
    """
    Tests if the tap installation failed
    """
    mock_failure = MagicMock(return_value={"stdout": "", "stderr": "", "retcode": 1})
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    with patch("salt.utils.path.which", MagicMock(return_value=HOMEBREW_BIN)):
        with patch.dict(
            mac_brew.__salt__,
            {
                "cmd.run_all": mock_failure,
                "file.get_user": mock_user,
                "cmd.run": mock_cmd,
            },
        ), patch("salt.modules.mac_brew_pkg._list_taps", MagicMock(return_value={})):
            assert not mac_brew._tap("homebrew/test")


def test_tap(TAPS_LIST, HOMEBREW_BIN):
    """
    Tests adding unofficial GitHub repos to the list of brew taps
    """
    mock_failure = MagicMock(return_value={"retcode": 0})
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    with patch("salt.utils.path.which", MagicMock(return_value=HOMEBREW_BIN)):
        with patch.dict(
            mac_brew.__salt__,
            {
                "cmd.run_all": mock_failure,
                "file.get_user": mock_user,
                "cmd.run": mock_cmd,
            },
        ), patch(
            "salt.modules.mac_brew_pkg._list_taps", MagicMock(return_value=TAPS_LIST)
        ):
            assert mac_brew._tap("homebrew/test")


# '_homebrew_bin' function tests: 1


def test_homebrew_bin(HOMEBREW_BIN):
    """
    Tests the path to the homebrew binary
    """
    mock_path = MagicMock(return_value="/usr/local")
    with patch("salt.utils.path.which", MagicMock(return_value=HOMEBREW_BIN)):
        with patch.dict(mac_brew.__salt__, {"cmd.run": mock_path}):
            assert mac_brew._homebrew_bin() == HOMEBREW_BIN


# 'list_pkgs' function tests: 2
# Only tested a few basics
# Full functionality should be tested in integration phase


def test_list_pkgs_removed():
    """
    Tests removed implementation
    """
    assert mac_brew.list_pkgs(removed=True) == {}


def test_list_pkgs_versions_true():
    """
    Tests if pkg.list_pkgs is already in context and is a list
    """
    mock_context = {"foo": ["bar"]}
    with patch.dict(mac_brew.__context__, {"pkg.list_pkgs": mock_context}):
        assert mac_brew.list_pkgs(versions_as_list=True) == mock_context


def test_list_pkgs_homebrew_cask_pakages():
    """
    Tests if pkg.list_pkgs list properly homebrew cask packages
    """
    expected_pkgs = {
        "homebrew/cask/day-o": "3.0.1",
        "day-o": "3.0.1",
        "homebrew/cask/discord": "0.0.266",
        "discord": "0.0.266",
        "custom/tap/iterm2": "3.4.3",
        "iterm2": "3.4.3",
        "jq": "1.6",
        "xz": "5.2.5",
    }

    with patch("salt.modules.mac_brew_pkg._call_brew", custom_call_brew), patch.dict(
        mac_brew.__salt__,
        {
            "pkg_resource.add_pkg": custom_add_pkg,
            "pkg_resource.sort_pkglist": MagicMock(),
        },
    ):
        assert mac_brew.list_pkgs(versions_as_list=True) == expected_pkgs


def test_list_pkgs_no_context():
    """
    Tests removed implementation
    """

    expected_pkgs = {
        "zsh": "5.7.1",
        "homebrew/cask/macvim": "8.1.151",
        "homebrew/cask-fonts/font-firacode-nerd-font": "2.0.0",
    }

    with patch("salt.modules.mac_brew_pkg._call_brew", custom_call_brew), patch.dict(
        mac_brew.__salt__,
        {
            "pkg_resource.add_pkg": custom_add_pkg,
            "pkg_resource.sort_pkglist": MagicMock(),
        },
    ), patch.object(mac_brew, "_list_pkgs_from_context") as list_pkgs_context_mock:
        pkgs = mac_brew.list_pkgs(versions_as_list=True, use_context=False)
        list_pkgs_context_mock.assert_not_called()
        list_pkgs_context_mock.reset_mock()

        pkgs = mac_brew.list_pkgs(versions_as_list=True, use_context=False)
        list_pkgs_context_mock.assert_not_called()
        list_pkgs_context_mock.reset_mock()


# 'version' function tests: 1


def test_version():
    """
    Tests version name returned
    """
    mock_version = MagicMock(return_value="0.1.5")
    with patch.dict(mac_brew.__salt__, {"pkg_resource.version": mock_version}):
        assert mac_brew.version("foo") == "0.1.5"


# 'latest_version' function tests: 0
# It has not been fully implemented

# 'remove' function tests: 1
# Only tested a few basics
# Full functionality should be tested in integration phase


def test_remove():
    """
    Tests if package to be removed exists
    """
    mock_params = MagicMock(return_value=({"foo": None}, "repository"))
    with patch(
        "salt.modules.mac_brew_pkg.list_pkgs", return_value={"test": "0.1.5"}
    ), patch.dict(mac_brew.__salt__, {"pkg_resource.parse_targets": mock_params}):
        assert mac_brew.remove("foo") == {}


# 'refresh_db' function tests: 2


def test_refresh_db_failure(HOMEBREW_BIN):
    """
    Tests an update of homebrew package repository failure
    """
    mock_user = MagicMock(return_value="foo")
    mock_failure = MagicMock(return_value={"stdout": "", "stderr": "", "retcode": 1})
    with patch("salt.utils.path.which", MagicMock(return_value="/usr/local/bin/brew")):
        with patch.dict(
            mac_brew.__salt__, {"file.get_user": mock_user, "cmd.run_all": mock_failure}
        ), patch(
            "salt.modules.mac_brew_pkg._homebrew_bin",
            MagicMock(return_value=HOMEBREW_BIN),
        ):
            with patch.object(salt.utils.pkg, "clear_rtag", Mock()):
                pytest.raises(CommandExecutionError, mac_brew.refresh_db)


def test_refresh_db(HOMEBREW_BIN):
    """
    Tests a successful update of homebrew package repository
    """
    mock_user = MagicMock(return_value="foo")
    mock_success = MagicMock(return_value={"retcode": 0})
    with patch("salt.utils.path.which", MagicMock(return_value=HOMEBREW_BIN)):
        with patch.dict(
            mac_brew.__salt__, {"file.get_user": mock_user, "cmd.run_all": mock_success}
        ), patch(
            "salt.modules.mac_brew_pkg._homebrew_bin",
            MagicMock(return_value=HOMEBREW_BIN),
        ):
            with patch.object(salt.utils.pkg, "clear_rtag", Mock()):
                assert mac_brew.refresh_db()


# 'install' function tests: 1
# Only tested a few basics
# Full functionality should be tested in integration phase


def test_install():
    """
    Tests if package to be installed exists
    """
    mock_params = MagicMock(return_value=[None, None])
    with patch.dict(mac_brew.__salt__, {"pkg_resource.parse_targets": mock_params}):
        assert mac_brew.install("name=foo") == {}


# "hold" function tests: 2
# Only tested a few basics
# Full functionality should be tested in integration phase


def test_hold(HOMEBREW_BIN):
    """
    Tests holding if package is installed
    """
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    mock_cmd_all = MagicMock(
        return_value={"pid": 12345, "retcode": 0, "stderr": "", "stdout": ""}
    )
    _expected = {
        "foo": {
            "changes": {"new": "hold", "old": "install"},
            "comment": "Package foo is now being held.",
            "name": "foo",
            "result": True,
        }
    }

    mock_params = MagicMock(return_value=({"foo": None}, "repository"))
    with patch("salt.utils.path.which", MagicMock(return_value=HOMEBREW_BIN)):
        with patch(
            "salt.modules.mac_brew_pkg.list_pkgs", return_value={"foo": "0.1.5"}
        ), patch.dict(
            mac_brew.__salt__,
            {
                "file.get_user": mock_user,
                "pkg_resource.parse_targets": mock_params,
                "cmd.run_all": mock_cmd_all,
                "cmd.run": mock_cmd,
            },
        ):
            assert mac_brew.hold("foo") == _expected


def test_hold_not_installed(HOMEBREW_BIN):
    """
    Tests holding if package is not installed
    """
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    mock_cmd_all = MagicMock(
        return_value={"pid": 12345, "retcode": 0, "stderr": "", "stdout": ""}
    )
    _expected = {
        "foo": {
            "changes": {},
            "comment": "Package foo does not have a state.",
            "name": "foo",
            "result": False,
        }
    }

    mock_params = MagicMock(return_value=({"foo": None}, "repository"))
    with patch("salt.utils.path.which", MagicMock(return_value=HOMEBREW_BIN)):
        with patch("salt.modules.mac_brew_pkg.list_pkgs", return_value={}), patch.dict(
            mac_brew.__salt__,
            {
                "file.get_user": mock_user,
                "pkg_resource.parse_targets": mock_params,
                "cmd.run_all": mock_cmd_all,
                "cmd.run": mock_cmd,
            },
        ):
            assert mac_brew.hold("foo") == _expected


def test_hold_pinned():
    """
    Tests holding if package is already pinned
    """
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    mock_cmd_all = MagicMock(
        return_value={"pid": 12345, "retcode": 0, "stderr": "", "stdout": ""}
    )
    _expected = {
        "foo": {
            "changes": {},
            "comment": "Package foo is already set to be held.",
            "name": "foo",
            "result": True,
        }
    }

    mock_params = MagicMock(return_value=({"foo": None}, "repository"))
    with patch(
        "salt.modules.mac_brew_pkg.list_pkgs", return_value={"foo": "0.1.5"}
    ), patch(
        "salt.modules.mac_brew_pkg._list_pinned", return_value=["foo"]
    ), patch.dict(
        mac_brew.__salt__,
        {
            "file.get_user": mock_user,
            "pkg_resource.parse_targets": mock_params,
            "cmd.run_all": mock_cmd_all,
            "cmd.run": mock_cmd,
        },
    ):
        assert mac_brew.hold("foo") == _expected


# "unhold" function tests: 2
# Only tested a few basics
# Full functionality should be tested in integration phase
def test_unhold(HOMEBREW_BIN):
    """
    Tests unholding if package is installed
    """
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    mock_cmd_all = MagicMock(
        return_value={"pid": 12345, "retcode": 0, "stderr": "", "stdout": ""}
    )
    _expected = {
        "foo": {
            "changes": {"new": "install", "old": "hold"},
            "comment": "Package foo is no longer being held.",
            "name": "foo",
            "result": True,
        }
    }

    mock_params = MagicMock(return_value=({"foo": None}, "repository"))
    with patch("salt.utils.path.which", MagicMock(return_value=HOMEBREW_BIN)):
        with patch(
            "salt.modules.mac_brew_pkg.list_pkgs", return_value={"foo": "0.1.5"}
        ), patch(
            "salt.modules.mac_brew_pkg._list_pinned", return_value=["foo"]
        ), patch.dict(
            mac_brew.__salt__,
            {
                "file.get_user": mock_user,
                "pkg_resource.parse_targets": mock_params,
                "cmd.run_all": mock_cmd_all,
                "cmd.run": mock_cmd,
            },
        ):
            assert mac_brew.unhold("foo") == _expected


def test_unhold_not_installed():
    """
    Tests unholding if package is not installed
    """
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    mock_cmd_all = MagicMock(
        return_value={"pid": 12345, "retcode": 0, "stderr": "", "stdout": ""}
    )
    _expected = {
        "foo": {
            "changes": {},
            "comment": "Package foo does not have a state.",
            "name": "foo",
            "result": False,
        }
    }

    mock_params = MagicMock(return_value=({"foo": None}, "repository"))
    with patch("salt.modules.mac_brew_pkg.list_pkgs", return_value={}), patch(
        "salt.modules.mac_brew_pkg._list_pinned", return_value=["foo"]
    ), patch.dict(
        mac_brew.__salt__,
        {
            "file.get_user": mock_user,
            "pkg_resource.parse_targets": mock_params,
            "cmd.run_all": mock_cmd_all,
            "cmd.run": mock_cmd,
        },
    ):
        assert mac_brew.unhold("foo") == _expected


def test_unhold_not_pinned():
    """
    Tests unholding if package is not installed
    """
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    mock_cmd_all = MagicMock(
        return_value={"pid": 12345, "retcode": 0, "stderr": "", "stdout": ""}
    )
    _expected = {
        "foo": {
            "changes": {},
            "comment": "Package foo is already set not to be held.",
            "name": "foo",
            "result": True,
        }
    }

    mock_params = MagicMock(return_value=({"foo": None}, "repository"))
    with patch(
        "salt.modules.mac_brew_pkg.list_pkgs", return_value={"foo": "0.1.5"}
    ), patch("salt.modules.mac_brew_pkg._list_pinned", return_value=[]), patch.dict(
        mac_brew.__salt__,
        {
            "file.get_user": mock_user,
            "pkg_resource.parse_targets": mock_params,
            "cmd.run_all": mock_cmd_all,
            "cmd.run": mock_cmd,
        },
    ):
        assert mac_brew.unhold("foo") == _expected


def test_info_installed(HOMEBREW_BIN):
    """
    Tests info_installed method
    """
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    mock_cmd_all = MagicMock(
        return_value={
            "pid": 12345,
            "retcode": 0,
            "stderr": "",
            "stdout": textwrap.dedent(
                """\
                {
                  "formulae": [
                    {
                      "name": "salt",
                      "full_name": "cdalvaro/tap/salt",
                      "tap": "cdalvaro/tap",
                      "aliases": []
                    },
                    {
                      "name": "vim",
                      "full_name": "vim",
                      "tap": "homebrew/core",
                      "aliases": []
                    }
                  ],
                  "casks": [
                    {
                      "token": "visual-studio-code",
                      "full_token": "visual-studio-code",
                      "tap": null,
                      "name": [
                        "MicrosoftVisualStudioCode",
                        "VSCode"
                      ]
                    }
                  ]
                }
             """
            ),
        }
    )
    _expected = {
        "cdalvaro/tap/salt": {
            "name": "salt",
            "full_name": "cdalvaro/tap/salt",
            "tap": "cdalvaro/tap",
            "aliases": [],
        },
        "vim": {
            "name": "vim",
            "full_name": "vim",
            "tap": "homebrew/core",
            "aliases": [],
        },
        "visual-studio-code": {
            "token": "visual-studio-code",
            "full_token": "visual-studio-code",
            "tap": None,
            "name": ["MicrosoftVisualStudioCode", "VSCode"],
        },
    }

    with patch("salt.utils.path.which", MagicMock(return_value=HOMEBREW_BIN)):
        with patch("salt.modules.mac_brew_pkg.list_pkgs", return_value={}), patch(
            "salt.modules.mac_brew_pkg._list_pinned", return_value=["foo"]
        ), patch.dict(
            mac_brew.__salt__,
            {
                "file.get_user": mock_user,
                "cmd.run_all": mock_cmd_all,
                "cmd.run": mock_cmd,
            },
        ):
            assert (
                mac_brew.info_installed(
                    "cdalvaro/tap/salt", "vim", "visual-studio-code"
                )
                == _expected
            )


def test_list_upgrades(HOMEBREW_BIN):
    """
    Tests list_upgrades method
    """
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    mock_cmd_all = MagicMock(
        return_value={
            "pid": 12345,
            "retcode": 0,
            "stderr": "",
            "stdout": textwrap.dedent(
                """\
                {
                  "formulae": [
                    {
                      "name": "cmake",
                      "installed_versions": ["3.19.3"],
                      "current_version": "3.19.4",
                      "pinned": false,
                      "pinned_version": null
                    },
                    {
                      "name": "fzf",
                      "installed_versions": ["0.25.0"],
                      "current_version": "0.25.1",
                      "pinned": false,
                      "pinned_version": null
                    }
                  ],
                  "casks": [
                    {
                      "name": "ksdiff",
                      "installed_versions": "2.2.0,122",
                      "current_version": "2.3.6,123-jan-18-2021"
                    }
                  ]
                }
                """
            ),
        }
    )
    _expected = {
        "cmake": "3.19.4",
        "fzf": "0.25.1",
        "ksdiff": "2.3.6,123-jan-18-2021",
    }

    with patch("salt.utils.path.which", MagicMock(return_value=HOMEBREW_BIN)):
        with patch("salt.modules.mac_brew_pkg.list_pkgs", return_value={}), patch(
            "salt.modules.mac_brew_pkg._list_pinned", return_value=["foo"]
        ), patch.dict(
            mac_brew.__salt__,
            {
                "file.get_user": mock_user,
                "cmd.run_all": mock_cmd_all,
                "cmd.run": mock_cmd,
            },
        ):
            assert (
                mac_brew.list_upgrades(refresh=False, include_casks=True) == _expected
            )



"""
Installing of certificates to the keychain
==========================================

Install certificats to the macOS keychain

.. code-block:: yaml

    /mnt/test.p12:
      keychain.installed:
        - password: test123
"""

import logging
import os

import salt.utils.platform

log = logging.getLogger(__name__)

__virtualname__ = "keychain"


def __virtual__():
    """
    Only work on Mac OS
    """
    if salt.utils.platform.is_darwin():
        return __virtualname__
    return (False, "Only supported on Mac OS")


def installed(name, password, keychain="/Library/Keychains/System.keychain", **kwargs):
    """
    Install a p12 certificate file into the macOS keychain

    name
        The certificate to install

    password
        The password for the certificate being installed formatted in the way
        described for openssl command in the PASS PHRASE ARGUMENTS section

    keychain
        The keychain to install the certificate to, this defaults to
        /Library/Keychains/System.keychain

    allow_any
        Allow any application to access the imported certificate without warning

    keychain_password
        If your keychain is likely to be locked pass the password and it will be unlocked
        before running the import

    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    if "http" in name or "salt" in name:
        name = __salt__["cp.cache_file"](name)

    certs = __salt__["keychain.list_certs"](keychain)
    friendly_name = __salt__["keychain.get_friendly_name"](name, password)

    if friendly_name in certs:
        file_hash = __salt__["keychain.get_hash"](name, password)
        keychain_hash = __salt__["keychain.get_hash"](friendly_name)

        if file_hash != keychain_hash:
            out = __salt__["keychain.uninstall"](
                friendly_name,
                keychain,
                keychain_password=kwargs.get("keychain_password"),
            )
            if "unable" not in out:
                ret["comment"] += (
                    "Found a certificate with the same name but different hash,"
                    " removing it.\n"
                )
                ret["changes"]["uninstalled"] = friendly_name

                # Reset the certs found
                certs = __salt__["keychain.list_certs"](keychain)
            else:
                ret["result"] = False
                ret[
                    "comment"
                ] += "Found an incorrect cert but was unable to uninstall it: {}".format(
                    friendly_name
                )
                return ret

    if friendly_name not in certs:
        out = __salt__["keychain.install"](name, password, keychain, **kwargs)
        if "imported" in out:
            ret["changes"]["installed"] = friendly_name
        else:
            ret["result"] = False
            ret["comment"] += "Failed to install {}".format(friendly_name)
    else:
        ret["comment"] += "{} already installed.".format(friendly_name)

    return ret


def uninstalled(
    name,
    password,
    keychain="/Library/Keychains/System.keychain",
    keychain_password=None,
):
    """
    Uninstall a p12 certificate file from the macOS keychain

    name
        The certificate to uninstall, this can be a path for a .p12 or the friendly
        name

    password
        The password for the certificate being installed formatted in the way
        described for openssl command in the PASS PHRASE ARGUMENTS section

    cert_name
        The friendly name of the certificate, this can be used instead of giving a
        certificate

    keychain
        The keychain to remove the certificate from, this defaults to
        /Library/Keychains/System.keychain

    keychain_password
        If your keychain is likely to be locked pass the password and it will be unlocked
        before running the import

    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    certs = __salt__["keychain.list_certs"](keychain)

    if ".p12" in name:
        if "http" in name or "salt" in name:
            name = __salt__["cp.cache_file"](name)

        friendly_name = __salt__["keychain.get_friendly_name"](name, password)
    else:
        friendly_name = name

    if friendly_name in certs:
        out = __salt__["keychain.uninstall"](friendly_name, keychain, keychain_password)
        if "unable" not in out:
            ret["changes"]["uninstalled"] = friendly_name
        else:
            ret["result"] = False
            ret["comment"] += "Failed to uninstall {}".format(friendly_name)
    else:
        ret["comment"] += "{} already uninstalled.".format(friendly_name)

    return ret


def default_keychain(name, domain="user", user=None):
    """
    Set the default keychain to use

    name
        The chain in which to use as the default

    domain
        The domain to use valid values are user|system|common|dynamic, the default is user

    user
        The user to run as

    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    if not os.path.exists(name):
        ret["result"] = False
        ret["comment"] += "Keychain not found at {}".format(name)
    else:
        out = __salt__["keychain.get_default_keychain"](user, domain)

        if name in out:
            ret["comment"] += "{} was already the default keychain.".format(name)
        else:
            out = __salt__["keychain.set_default_keychain"](name, domain, user)
            if len(out) == 0:
                ret["changes"]["default"] = name
            else:
                ret["result"] = False
                ret["comment"] = "Failed to install keychain. {}".format(out)

    return ret



"""
Manage Grafana v4.0 data sources

.. versionadded:: 2017.7.0

:configuration: This state requires a configuration profile to be configured
    in the minion config, minion pillar, or master config. The module will use
    the 'grafana' key by default, if defined.

    Example configuration using basic authentication:

    .. code-block:: yaml

        grafana:
          grafana_url: http://grafana.localhost
          grafana_user: admin
          grafana_password: admin
          grafana_timeout: 3

    Example configuration using token based authentication:

    .. code-block:: yaml

        grafana:
          grafana_url: http://grafana.localhost
          grafana_token: token
          grafana_timeout: 3

The behavior of this module is to create data sources if the do not exists, and
to update data sources if the already exists.

.. code-block:: yaml

    Ensure influxdb data source is present:
      grafana4_datasource.present:
        - name: influxdb
        - type: influxdb
        - url: http://localhost:8086
        - access: proxy
        - basic_auth: true
        - basic_auth_user: myuser
        - basic_auth_password: mypass
        - is_default: true
"""

from salt.utils.dictdiffer import deep_diff


def __virtual__():
    """Only load if grafana4 module is available"""
    if "grafana4.get_datasource" in __salt__:
        return True
    return (False, "grafana4 module could not be loaded")


def present(
    name,
    type,
    url,
    access=None,
    user=None,
    password=None,
    database=None,
    basic_auth=None,
    basic_auth_user=None,
    basic_auth_password=None,
    tls_auth=None,
    json_data=None,
    is_default=None,
    with_credentials=None,
    type_logo_url=None,
    orgname=None,
    profile="grafana",
):
    """
    Ensure that a data source is present.

    name
        Name of the data source.

    type
        Type of the datasource ('graphite', 'influxdb' etc.).

    access
        Use proxy or direct. Default: proxy

    url
        The URL to the data source API.

    user
        Optional - user to authenticate with the data source.

    password
        Optional - password to authenticate with the data source.

    database
        Optional - database to use with the data source.

    basic_auth
        Optional - set to True to use HTTP basic auth to authenticate with the
        data source.

    basic_auth_user
        Optional - HTTP basic auth username.

    basic_auth_password
        Optional - HTTP basic auth password.

    json_data
        Optional - additional json data to post (eg. "timeInterval").

    is_default
        Optional - set data source as default.

    with_credentials
        Optional - Whether credentials such as cookies or auth headers should
        be sent with cross-site requests.

    type_logo_url
        Optional - Logo to use for this datasource.

    orgname
        Name of the organization in which the data source should be present.

    profile
        Configuration profile used to connect to the Grafana instance.
        Default is 'grafana'.
    """
    if isinstance(profile, str):
        profile = __salt__["config.option"](profile)

    ret = {"name": name, "result": None, "comment": None, "changes": {}}
    datasource = __salt__["grafana4.get_datasource"](name, orgname, profile)
    data = _get_json_data(
        name=name,
        type=type,
        url=url,
        access=access,
        user=user,
        password=password,
        database=database,
        basicAuth=basic_auth,
        basicAuthUser=basic_auth_user,
        basicAuthPassword=basic_auth_password,
        tlsAuth=tls_auth,
        jsonData=json_data,
        isDefault=is_default,
        withCredentials=with_credentials,
        typeLogoUrl=type_logo_url,
        defaults=datasource,
    )

    if not datasource:
        if __opts__["test"]:
            ret["comment"] = "Datasource {} will be created".format(name)
            return ret
        __salt__["grafana4.create_datasource"](profile=profile, **data)
        datasource = __salt__["grafana4.get_datasource"](name, profile=profile)
        ret["result"] = True
        ret["comment"] = "New data source {} added".format(name)
        ret["changes"] = data
        return ret

    # At this stage, the datasource exists; however, the object provided by
    # Grafana may lack some null keys compared to our "data" dict:
    for key in data:
        if key not in datasource:
            datasource[key] = None

    if data == datasource:
        ret["comment"] = "Data source {} already up-to-date".format(name)
        return ret

    if __opts__["test"]:
        ret["comment"] = "Datasource {} will be updated".format(name)
        return ret
    __salt__["grafana4.update_datasource"](datasource["id"], profile=profile, **data)
    ret["result"] = True
    ret["changes"] = deep_diff(datasource, data, ignore=["id", "orgId", "readOnly"])
    ret["comment"] = "Data source {} updated".format(name)
    return ret


def absent(name, orgname=None, profile="grafana"):
    """
    Ensure that a data source is present.

    name
        Name of the data source to remove.

    orgname
        Name of the organization from which the data source should be absent.

    profile
        Configuration profile used to connect to the Grafana instance.
        Default is 'grafana'.
    """
    if isinstance(profile, str):
        profile = __salt__["config.option"](profile)

    ret = {"name": name, "result": None, "comment": None, "changes": {}}
    datasource = __salt__["grafana4.get_datasource"](name, orgname, profile)

    if not datasource:
        ret["result"] = True
        ret["comment"] = "Data source {} already absent".format(name)
        return ret

    if __opts__["test"]:
        ret["comment"] = "Datasource {} will be deleted".format(name)
        return ret
    __salt__["grafana4.delete_datasource"](datasource["id"], profile=profile)

    ret["result"] = True
    ret["changes"][name] = "Absent"
    ret["comment"] = "Data source {} was deleted".format(name)

    return ret


def _get_json_data(defaults=None, **kwargs):
    if defaults is None:
        defaults = {}
    for k, v in kwargs.items():
        if v is None:
            kwargs[k] = defaults.get(k)
    return kwargs



import pytest

import salt.modules.pdbedit as pdbedit_mod
import salt.states.pdbedit as pdbedit
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {pdbedit: {}, pdbedit_mod: {}}


def test_generate_absent():
    """
    Test salt.states.pdbedit.absent when
    user is already absent
    """
    name = "testname"
    cmd_ret = {"pid": 13172, "retcode": 0, "stdout": "", "stderr": ""}
    with patch.dict(pdbedit.__salt__, {"pdbedit.list": pdbedit_mod.list_users}):
        with patch.dict(
            pdbedit_mod.__salt__, {"cmd.run_all": MagicMock(return_value=cmd_ret)}
        ):
            ret = pdbedit.absent(name)
    assert ret["comment"] == "account {} is absent".format(name)



"""
    Test cases for salt.modules.keystone
"""


import pytest

import salt.modules.config as config
import salt.modules.zenoss as zenoss
from tests.support.mock import MagicMock, call, patch


@pytest.fixture
def configure_loader_modules():
    return {
        zenoss: {"__salt__": {"config.option": config.option}},
        config: {"__opts__": {}},
    }


def test_zenoss_session():
    """
    test zenoss._session when using verify_ssl
    """
    zenoss_conf = {
        "zenoss": {
            "hostname": "https://test.zenoss.com",
            "username": "admin",
            "password": "test123",
        }
    }

    for verify in [True, False, None]:
        zenoss_conf["zenoss"]["verify_ssl"] = verify
        if verify is None:
            zenoss_conf["zenoss"].pop("verify_ssl")
            verify = True

        patch_opts = patch.dict(config.__opts__, zenoss_conf)
        mock_http = MagicMock(return_value=None)
        patch_http = patch("salt.utils.http.session", mock_http)

        with patch_http, patch_opts:
            zenoss._session()
            assert mock_http.call_args_list == [
                call(
                    ca_bundle=None,
                    headers={"Content-type": "application/json; charset=utf-8"},
                    password="test123",
                    user="admin",
                    verify_ssl=verify,
                )
            ]



"""
Common resources for LXC and systemd-nspawn containers

.. versionadded:: 2015.8.0

These functions are not designed to be called directly, but instead from the
:mod:`lxc <salt.modules.lxc>`, :mod:`nspawn <salt.modules.nspawn>`, and
:mod:`docker <salt.modules.docker>` execution modules. They provide for
common logic to be re-used for common actions.
"""
import copy
import functools
import logging
import os
import shlex
import time
import traceback

import salt.utils.args
import salt.utils.path
import salt.utils.vt
from salt.exceptions import CommandExecutionError, SaltInvocationError

log = logging.getLogger(__name__)

PATH = "PATH=/bin:/usr/bin:/sbin:/usr/sbin:/opt/bin:/usr/local/bin:/usr/local/sbin"


def _validate(wrapped):
    """
    Decorator for common function argument validation
    """

    @functools.wraps(wrapped)
    def wrapper(*args, **kwargs):
        container_type = kwargs.get("container_type")
        exec_driver = kwargs.get("exec_driver")
        valid_driver = {
            "docker": ("lxc-attach", "nsenter", "docker-exec"),
            "lxc": ("lxc-attach",),
            "nspawn": ("nsenter",),
        }
        if container_type not in valid_driver:
            raise SaltInvocationError(
                "Invalid container type '{}'. Valid types are: {}".format(
                    container_type, ", ".join(sorted(valid_driver))
                )
            )
        if exec_driver not in valid_driver[container_type]:
            raise SaltInvocationError(
                "Invalid command execution driver. Valid drivers are: {}".format(
                    ", ".join(valid_driver[container_type])
                )
            )
        if exec_driver == "lxc-attach" and not salt.utils.path.which("lxc-attach"):
            raise SaltInvocationError(
                "The 'lxc-attach' execution driver has been chosen, but "
                "lxc-attach is not available. LXC may not be installed."
            )
        return wrapped(*args, **salt.utils.args.clean_kwargs(**kwargs))

    return wrapper


def _nsenter(pid):
    """
    Return the nsenter command to attach to the named container
    """
    return f"nsenter --target {pid} --mount --uts --ipc --net --pid"


def _get_md5(name, path, run_func):
    """
    Get the MD5 checksum of a file from a container
    """
    output = run_func(name, f"md5sum {shlex.quote(path)}", ignore_retcode=True)[
        "stdout"
    ]
    try:
        return output.split()[0]
    except IndexError:
        # Destination file does not exist or could not be accessed
        return None


def cache_file(source):
    """
    Wrapper for cp.cache_file which raises an error if the file was unable to
    be cached.

    CLI Example:

    .. code-block:: bash

        salt myminion container_resource.cache_file salt://foo/bar/baz.txt
    """
    try:
        # Don't just use cp.cache_file for this. Docker has its own code to
        # pull down images from the web.
        if source.startswith("salt://"):
            cached_source = __salt__["cp.cache_file"](source)
            if not cached_source:
                raise CommandExecutionError(f"Unable to cache {source}")
            return cached_source
    except AttributeError:
        raise SaltInvocationError(f"Invalid source file {source}")
    return source


@_validate
def run(
    name,
    cmd,
    container_type=None,
    exec_driver=None,
    output=None,
    no_start=False,
    stdin=None,
    python_shell=True,
    output_loglevel="debug",
    ignore_retcode=False,
    path=None,
    use_vt=False,
    keep_env=None,
):
    """
    Common logic for running shell commands in containers

    path
        path to the container parent (for LXC only)
        default: /var/lib/lxc (system default)

    CLI Example:

    .. code-block:: bash

        salt myminion container_resource.run mycontainer 'ps aux' container_type=docker exec_driver=nsenter output=stdout
    """
    valid_output = ("stdout", "stderr", "retcode", "all")
    if output is None:
        cmd_func = "cmd.run"
    elif output not in valid_output:
        raise SaltInvocationError(
            "'output' param must be one of the following: {}".format(
                ", ".join(valid_output)
            )
        )
    else:
        cmd_func = "cmd.run_all"

    if keep_env is None or isinstance(keep_env, bool):
        to_keep = []
    elif not isinstance(keep_env, (list, tuple)):
        try:
            to_keep = keep_env.split(",")
        except AttributeError:
            log.warning("Invalid keep_env value, ignoring")
            to_keep = []
    else:
        to_keep = keep_env

    if exec_driver == "lxc-attach":
        full_cmd = "lxc-attach "
        if path:
            full_cmd += f"-P {shlex.quote(path)} "
        if keep_env is not True:
            full_cmd += "--clear-env "
            if "PATH" not in to_keep:
                full_cmd += f"--set-var {PATH} "
                # --clear-env results in a very restrictive PATH
                # (/bin:/usr/bin), use a good fallback.
        full_cmd += " ".join(
            [
                f"--set-var {x}={shlex.quote(os.environ[x])}"
                for x in to_keep
                if x in os.environ
            ]
        )
        full_cmd += f" -n {shlex.quote(name)} -- {cmd}"
    elif exec_driver == "nsenter":
        pid = __salt__[f"{container_type}.pid"](name)
        full_cmd = f"nsenter --target {pid} --mount --uts --ipc --net --pid -- "
        if keep_env is not True:
            full_cmd += "env -i "
            if "PATH" not in to_keep:
                full_cmd += f"{PATH} "
        full_cmd += " ".join(
            [f"{x}={shlex.quote(os.environ[x])}" for x in to_keep if x in os.environ]
        )
        full_cmd += f" {cmd}"
    elif exec_driver == "docker-exec":
        # We're using docker exec on the CLI as opposed to via docker-py, since
        # the Docker API doesn't return stdout and stderr separately.
        full_cmd = "docker exec "
        if stdin:
            full_cmd += "-i "
        full_cmd += f"{name} "
        if keep_env is not True:
            full_cmd += "env -i "
            if "PATH" not in to_keep:
                full_cmd += f"{PATH} "
        full_cmd += " ".join(
            [f"{x}={shlex.quote(os.environ[x])}" for x in to_keep if x in os.environ]
        )
        full_cmd += f" {cmd}"

    if not use_vt:
        ret = __salt__[cmd_func](
            full_cmd,
            stdin=stdin,
            python_shell=python_shell,
            output_loglevel=output_loglevel,
            ignore_retcode=ignore_retcode,
        )
    else:
        stdout, stderr = "", ""
        proc = salt.utils.vt.Terminal(
            full_cmd,
            shell=python_shell,
            log_stdin_level="quiet" if output_loglevel == "quiet" else "info",
            log_stdout_level=output_loglevel,
            log_stderr_level=output_loglevel,
            log_stdout=True,
            log_stderr=True,
            stream_stdout=False,
            stream_stderr=False,
        )
        # Consume output
        try:
            while proc.has_unread_data:
                try:
                    cstdout, cstderr = proc.recv()
                    if cstdout:
                        stdout += cstdout
                    if cstderr:
                        if output is None:
                            stdout += cstderr
                        else:
                            stderr += cstderr
                    time.sleep(0.5)
                except KeyboardInterrupt:
                    break
            ret = (
                stdout
                if output is None
                else {
                    "retcode": proc.exitstatus,
                    "pid": 2,
                    "stdout": stdout,
                    "stderr": stderr,
                }
            )
        except salt.utils.vt.TerminalException:
            trace = traceback.format_exc()
            log.error(trace)
            ret = (
                stdout
                if output is None
                else {"retcode": 127, "pid": 2, "stdout": stdout, "stderr": stderr}
            )
        finally:
            proc.terminate()

    return ret


@_validate
def copy_to(
    name,
    source,
    dest,
    container_type=None,
    path=None,
    exec_driver=None,
    overwrite=False,
    makedirs=False,
):
    """
    Common logic for copying files to containers

    path
        path to the container parent (for LXC only)
        default: /var/lib/lxc (system default)

    CLI Example:

    .. code-block:: bash

        salt myminion container_resource.copy_to mycontainer /local/file/path /container/file/path container_type=docker exec_driver=nsenter
    """
    # Get the appropriate functions
    state = __salt__[f"{container_type}.state"]

    def run_all(*args, **akwargs):
        akwargs = copy.deepcopy(akwargs)
        if container_type in ["lxc"] and "path" not in akwargs:
            akwargs["path"] = path
        return __salt__[f"{container_type}.run_all"](*args, **akwargs)

    state_kwargs = {}
    cmd_kwargs = {"ignore_retcode": True}
    if container_type in ["lxc"]:
        cmd_kwargs["path"] = path
        state_kwargs["path"] = path

    def _state(name):
        if state_kwargs:
            return state(name, **state_kwargs)
        else:
            return state(name)

    c_state = _state(name)
    if c_state != "running":
        raise CommandExecutionError(f"Container '{name}' is not running")

    local_file = cache_file(source)
    source_dir, source_name = os.path.split(local_file)

    # Source file sanity checks
    if not os.path.isabs(local_file):
        raise SaltInvocationError("Source path must be absolute")
    elif not os.path.exists(local_file):
        raise SaltInvocationError(f"Source file {local_file} does not exist")
    elif not os.path.isfile(local_file):
        raise SaltInvocationError("Source must be a regular file")

    # Destination file sanity checks
    if not os.path.isabs(dest):
        raise SaltInvocationError("Destination path must be absolute")
    if run_all(name, f"test -d {shlex.quote(dest)}", **cmd_kwargs)["retcode"] == 0:
        # Destination is a directory, full path to dest file will include the
        # basename of the source file.
        dest = os.path.join(dest, source_name)
    else:
        # Destination was not a directory. We will check to see if the parent
        # dir is a directory, and then (if makedirs=True) attempt to create the
        # parent directory.
        dest_dir, dest_name = os.path.split(dest)
        if (
            run_all(name, f"test -d {shlex.quote(dest_dir)}", **cmd_kwargs)["retcode"]
            != 0
        ):
            if makedirs:
                result = run_all(
                    name, f"mkdir -p {shlex.quote(dest_dir)}", **cmd_kwargs
                )
                if result["retcode"] != 0:
                    error = (
                        "Unable to create destination directory {} in "
                        "container '{}'".format(dest_dir, name)
                    )
                    if result["stderr"]:
                        error += ": {}".format(result["stderr"])
                    raise CommandExecutionError(error)
            else:
                raise SaltInvocationError(
                    "Directory {} does not exist on {} container '{}'".format(
                        dest_dir, container_type, name
                    )
                )
    if (
        not overwrite
        and run_all(name, f"test -e {shlex.quote(dest)}", **cmd_kwargs)["retcode"] == 0
    ):
        raise CommandExecutionError(
            "Destination path {} already exists. Use overwrite=True to "
            "overwrite it".format(dest)
        )

    # Before we try to replace the file, compare checksums.
    source_md5 = __salt__["file.get_sum"](local_file, "md5")
    if source_md5 == _get_md5(name, dest, run_all):
        log.debug("%s and %s:%s are the same file, skipping copy", source, name, dest)
        return True

    log.debug(
        "Copying %s to %s container '%s' as %s", source, container_type, name, dest
    )

    # Using cat here instead of opening the file, reading it into memory,
    # and passing it as stdin to run(). This will keep down memory
    # usage for the minion and make the operation run quicker.
    if exec_driver == "lxc-attach":
        lxcattach = "lxc-attach"
        if path:
            lxcattach += f" -P {shlex.quote(path)}"
        copy_cmd = (
            'cat "{0}" | {4} --clear-env --set-var {1} -n {2} -- tee "{3}"'.format(
                local_file, PATH, name, dest, lxcattach
            )
        )
    elif exec_driver == "nsenter":
        pid = __salt__[f"{container_type}.pid"](name)
        copy_cmd = 'cat "{}" | {} env -i {} tee "{}"'.format(
            local_file, _nsenter(pid), PATH, dest
        )
    elif exec_driver == "docker-exec":
        copy_cmd = 'cat "{}" | docker exec -i {} env -i {} tee "{}"'.format(
            local_file, name, PATH, dest
        )
    __salt__["cmd.run"](copy_cmd, python_shell=True, output_loglevel="quiet")
    return source_md5 == _get_md5(name, dest, run_all)



"""
Namecheap SSL Certificate Management

.. versionadded:: 2017.7.0

Prerequisites
-------------

This module uses the ``requests`` Python module to communicate to the namecheap
API.

Configuration
-------------

The Namecheap username, API key and URL should be set in the minion configuration
file, or in the Pillar data.

.. code-block:: yaml

    namecheap.name: companyname
    namecheap.key: a1b2c3d4e5f67a8b9c0d1e2f3
    namecheap.client_ip: 162.155.30.172
    #Real url
    namecheap.url: https://api.namecheap.com/xml.response
    #Sandbox url
    #namecheap.url: https://api.sandbox.namecheap.xml.response
"""

import logging

import salt.utils.files
import salt.utils.stringutils

try:
    import salt.utils.namecheap

    CAN_USE_NAMECHEAP = True
except ImportError:
    CAN_USE_NAMECHEAP = False


log = logging.getLogger(__name__)


def __virtual__():
    """
    Check to make sure requests and xml are installed and requests
    """
    if CAN_USE_NAMECHEAP:
        return "namecheap_ssl"
    return False


def reissue(
    csr_file,
    certificate_id,
    web_server_type,
    approver_email=None,
    http_dc_validation=False,
    **kwargs
):
    """
    Reissues a purchased SSL certificate. Returns a dictionary of result
    values.

    csr_file
        Path to Certificate Signing Request file

    certificate_id
        Unique ID of the SSL certificate you wish to activate

    web_server_type
        The type of certificate format to return. Possible values include:

        - apache2
        - apacheapachessl
        - apacheopenssl
        - apacheraven
        - apachessl
        - apachessleay
        - c2net
        - cobaltseries
        - cpanel
        - domino
        - dominogo4625
        - dominogo4626
        - ensim
        - hsphere
        - ibmhttp
        - iis
        - iis4
        - iis5
        - iplanet
        - ipswitch
        - netscape
        - other
        - plesk
        - tomcat
        - weblogic
        - website
        - webstar
        - zeusv3

    approver_email
        The email ID which is on the approver email list.

        .. note::
            ``http_dc_validation`` must be set to ``False`` if this option is
            used.

    http_dc_validation : False
        Whether or not to activate using HTTP-based validation.

    .. note::
        For other parameters which may be required, see here__.

        .. __: https://www.namecheap.com/support/api/methods/ssl/reissue.aspx

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' namecheap_ssl.reissue my-csr-file my-cert-id apachessl
    """
    return __get_certificates(
        "namecheap.ssl.reissue",
        "SSLReissueResult",
        csr_file,
        certificate_id,
        web_server_type,
        approver_email,
        http_dc_validation,
        kwargs,
    )


def activate(
    csr_file,
    certificate_id,
    web_server_type,
    approver_email=None,
    http_dc_validation=False,
    **kwargs
):
    """
    Activates a newly-purchased SSL certificate. Returns a dictionary of result
    values.

    csr_file
        Path to Certificate Signing Request file

    certificate_id
        Unique ID of the SSL certificate you wish to activate

    web_server_type
        The type of certificate format to return. Possible values include:

        - apache2
        - apacheapachessl
        - apacheopenssl
        - apacheraven
        - apachessl
        - apachessleay
        - c2net
        - cobaltseries
        - cpanel
        - domino
        - dominogo4625
        - dominogo4626
        - ensim
        - hsphere
        - ibmhttp
        - iis
        - iis4
        - iis5
        - iplanet
        - ipswitch
        - netscape
        - other
        - plesk
        - tomcat
        - weblogic
        - website
        - webstar
        - zeusv3

    approver_email
        The email ID which is on the approver email list.

        .. note::
            ``http_dc_validation`` must be set to ``False`` if this option is
            used.

    http_dc_validation : False
        Whether or not to activate using HTTP-based validation.

    .. note::
        For other parameters which may be required, see here__.

        .. __: https://www.namecheap.com/support/api/methods/ssl/activate.aspx

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' namecheap_ssl.activate my-csr-file my-cert-id apachessl
    """
    return __get_certificates(
        "namecheap.ssl.activate",
        "SSLActivateResult",
        csr_file,
        certificate_id,
        web_server_type,
        approver_email,
        http_dc_validation,
        kwargs,
    )


def __get_certificates(
    command,
    result_tag_name,
    csr_file,
    certificate_id,
    web_server_type,
    approver_email,
    http_dc_validation,
    kwargs,
):

    web_server_types = (
        "apacheopenssl",
        "apachessl",
        "apacheraven",
        "apachessleay",
        "c2net",
        "ibmhttp",
        "iplanet",
        "domino",
        "dominogo4625",
        "dominogo4626",
        "netscape",
        "zeusv3",
        "apache2",
        "apacheapachessl",
        "cobaltseries",
        "cpanel",
        "ensim",
        "hsphere",
        "ipswitch",
        "plesk",
        "tomcat",
        "weblogic",
        "website",
        "webstar",
        "iis",
        "other",
        "iis4",
        "iis5",
    )

    if web_server_type not in web_server_types:
        log.error("Invalid option for web_server_type=%s", web_server_type)
        raise Exception("Invalid option for web_server_type=" + web_server_type)

    if approver_email is not None and http_dc_validation:
        log.error("approver_email and http_dc_validation cannot both have values")
        raise Exception("approver_email and http_dc_validation cannot both have values")

    if approver_email is None and not http_dc_validation:
        log.error("approver_email or http_dc_validation must have a value")
        raise Exception("approver_email or http_dc_validation must have a value")

    opts = salt.utils.namecheap.get_opts(command)

    with salt.utils.files.fopen(csr_file, "rb") as csr_handle:
        opts["csr"] = salt.utils.stringutils.to_unicode(csr_handle.read())

    opts["CertificateID"] = certificate_id
    opts["WebServerType"] = web_server_type
    if approver_email is not None:
        opts["ApproverEmail"] = approver_email

    if http_dc_validation:
        opts["HTTPDCValidation"] = "True"

    for key, value in kwargs.items():
        opts[key] = value

    response_xml = salt.utils.namecheap.post_request(opts)

    if response_xml is None:
        return {}

    sslresult = response_xml.getElementsByTagName(result_tag_name)[0]
    result = salt.utils.namecheap.atts_to_dict(sslresult)

    if http_dc_validation:
        validation_tag = sslresult.getElementsByTagName("HttpDCValidation")
        if validation_tag:
            validation_tag = validation_tag[0]

            if validation_tag.getAttribute("ValueAvailable").lower() == "true":
                validation_dict = {
                    "filename": validation_tag.getElementsByTagName("FileName")[0]
                    .childNodes[0]
                    .data,
                    "filecontent": validation_tag.getElementsByTagName("FileContent")[0]
                    .childNodes[0]
                    .data,
                }
                result["httpdcvalidation"] = validation_dict

    return result


def renew(years, certificate_id, certificate_type, promotion_code=None):
    """
    Renews an SSL certificate if it is ACTIVE and Expires <= 30 days. Returns
    the following information:

    - The certificate ID
    - The order ID
    - The transaction ID
    - The amount charged for the order

    years : 1
        Number of years to register

    certificate_id
        Unique ID of the SSL certificate you wish to renew

    certificate_type
        Type of SSL Certificate. Possible values include:

        - EV Multi Domain SSL
        - EV SSL
        - EV SSL SGC
        - EssentialSSL
        - EssentialSSL Wildcard
        - InstantSSL
        - InstantSSL Pro
        - Multi Domain SSL
        - PositiveSSL
        - PositiveSSL Multi Domain
        - PositiveSSL Wildcard
        - PremiumSSL
        - PremiumSSL Wildcard
        - QuickSSL Premium
        - RapidSSL
        - RapidSSL Wildcard
        - SGC Supercert
        - SSL Web Server
        - SSL Webserver EV
        - SSL123
        - Secure Site
        - Secure Site Pro
        - Secure Site Pro with EV
        - Secure Site with EV
        - True BusinessID
        - True BusinessID Multi Domain
        - True BusinessID Wildcard
        - True BusinessID with EV
        - True BusinessID with EV Multi Domain
        - Unified Communications

    promotional_code
        An optional promo code to use when renewing the certificate

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' namecheap_ssl.renew 1 my-cert-id RapidSSL
    """

    valid_certs = (
        "QuickSSL Premium",
        "RapidSSL",
        "RapidSSL Wildcard",
        "PremiumSSL",
        "InstantSSL",
        "PositiveSSL",
        "PositiveSSL Wildcard",
        "True BusinessID with EV",
        "True BusinessID",
        "True BusinessID Wildcard",
        "True BusinessID Multi Domain",
        "True BusinessID with EV Multi Domain",
        "Secure Site",
        "Secure Site Pro",
        "Secure Site with EV",
        "Secure Site Pro with EV",
        "EssentialSSL",
        "EssentialSSL Wildcard",
        "InstantSSL Pro",
        "PremiumSSL Wildcard",
        "EV SSL",
        "EV SSL SGC",
        "SSL123",
        "SSL Web Server",
        "SGC Supercert",
        "SSL Webserver EV",
        "EV Multi Domain SSL",
        "Multi Domain SSL",
        "PositiveSSL Multi Domain",
        "Unified Communications",
    )

    if certificate_type not in valid_certs:
        log.error("Invalid option for certificate_type=%s", certificate_type)
        raise Exception("Invalid option for certificate_type=" + certificate_type)

    if years < 1 or years > 5:
        log.error("Invalid option for years=%s", str(years))
        raise Exception("Invalid option for years=" + str(years))

    opts = salt.utils.namecheap.get_opts("namecheap.ssl.renew")
    opts["Years"] = str(years)
    opts["CertificateID"] = str(certificate_id)
    opts["SSLType"] = certificate_type
    if promotion_code is not None:
        opts["PromotionCode"] = promotion_code

    response_xml = salt.utils.namecheap.post_request(opts)
    if response_xml is None:
        return {}

    sslrenewresult = response_xml.getElementsByTagName("SSLRenewResult")[0]
    return salt.utils.namecheap.atts_to_dict(sslrenewresult)


def create(years, certificate_type, promotion_code=None, sans_to_add=None):
    """
    Creates a new SSL certificate. Returns the following information:

    - Whether or not the SSL order was successful
    - The certificate ID
    - The order ID
    - The transaction ID
    - The amount charged for the order
    - The date on which the certificate was created
    - The date on which the certificate will expire
    - The type of SSL certificate
    - The number of years for which the certificate was purchased
    - The current status of the SSL certificate

    years : 1
        Number of years to register

    certificate_type
        Type of SSL Certificate. Possible values include:

        - EV Multi Domain SSL
        - EV SSL
        - EV SSL SGC
        - EssentialSSL
        - EssentialSSL Wildcard
        - InstantSSL
        - InstantSSL Pro
        - Multi Domain SSL
        - PositiveSSL
        - PositiveSSL Multi Domain
        - PositiveSSL Wildcard
        - PremiumSSL
        - PremiumSSL Wildcard
        - QuickSSL Premium
        - RapidSSL
        - RapidSSL Wildcard
        - SGC Supercert
        - SSL Web Server
        - SSL Webserver EV
        - SSL123
        - Secure Site
        - Secure Site Pro
        - Secure Site Pro with EV
        - Secure Site with EV
        - True BusinessID
        - True BusinessID Multi Domain
        - True BusinessID Wildcard
        - True BusinessID with EV
        - True BusinessID with EV Multi Domain
        - Unified Communications

    promotional_code
        An optional promo code to use when creating the certificate

    sans_to_add : 0
        This parameter defines the number of add-on domains to be purchased in
        addition to the default number of domains included with a multi-domain
        certificate. Each certificate that supports SANs has the default number
        of domains included. You may check the default number of domains
        included and the maximum number of domains that can be added to it in
        the table below.

    +----------+----------------+----------------------+-------------------+----------------+
    | Provider | Product name   | Default number of    | Maximum number of | Maximum number |
    |          |                | domains (domain from | total domains     | of domains     |
    |          |                | CSR is counted here) |                   | that can be    |
    |          |                |                      |                   | passed in      |
    |          |                |                      |                   | sans_to_add    |
    |          |                |                      |                   | parameter      |
    +----------+----------------+----------------------+-------------------+----------------+
    | Comodo   | PositiveSSL    | 3                    | 100               | 97             |
    |          | Multi-Domain   |                      |                   |                |
    +----------+----------------+----------------------+-------------------+----------------+
    | Comodo   | Multi-Domain   | 3                    | 100               | 97             |
    |          | SSL            |                      |                   |                |
    +----------+----------------+----------------------+-------------------+----------------+
    | Comodo   | EV Multi-      | 3                    | 100               | 97             |
    |          | Domain SSL     |                      |                   |                |
    +----------+----------------+----------------------+-------------------+----------------+
    | Comodo   | Unified        | 3                    | 100               | 97             |
    |          | Communications |                      |                   |                |
    +----------+----------------+----------------------+-------------------+----------------+
    | GeoTrust | QuickSSL       | 1                    | 1 domain +        | The only       |
    |          | Premium        |                      | 4 subdomains      | supported      |
    |          |                |                      |                   | value is 4     |
    +----------+----------------+----------------------+-------------------+----------------+
    | GeoTrust | True           | 5                    | 25                | 20             |
    |          | BusinessID     |                      |                   |                |
    |          | with EV        |                      |                   |                |
    |          | Multi-Domain   |                      |                   |                |
    +----------+----------------+----------------------+-------------------+----------------+
    | GeoTrust | True Business  | 5                    | 25                | 20             |
    |          | ID Multi-      |                      |                   |                |
    |          | Domain         |                      |                   |                |
    +----------+----------------+----------------------+-------------------+----------------+
    | Thawte   | SSL Web        | 1                    | 25                | 24             |
    |          | Server         |                      |                   |                |
    +----------+----------------+----------------------+-------------------+----------------+
    | Thawte   | SSL Web        | 1                    | 25                | 24             |
    |          | Server with    |                      |                   |                |
    |          | EV             |                      |                   |                |
    +----------+----------------+----------------------+-------------------+----------------+
    | Thawte   | SGC Supercerts | 1                    | 25                | 24             |
    +----------+----------------+----------------------+-------------------+----------------+
    | Symantec | Secure Site    | 1                    | 25                | 24             |
    |          | Pro with EV    |                      |                   |                |
    +----------+----------------+----------------------+-------------------+----------------+
    | Symantec | Secure Site    | 1                    | 25                | 24             |
    |          | with EV        |                      |                   |                |
    +----------+----------------+----------------------+-------------------+----------------+
    | Symantec | Secure Site    | 1                    | 25                | 24             |
    +----------+----------------+----------------------+-------------------+----------------+
    | Symantec | Secure Site    | 1                    | 25                | 24             |
    |          | Pro            |                      |                   |                |
    +----------+----------------+----------------------+-------------------+----------------+

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' namecheap_ssl.create 2 RapidSSL
    """
    valid_certs = (
        "QuickSSL Premium",
        "RapidSSL",
        "RapidSSL Wildcard",
        "PremiumSSL",
        "InstantSSL",
        "PositiveSSL",
        "PositiveSSL Wildcard",
        "True BusinessID with EV",
        "True BusinessID",
        "True BusinessID Wildcard",
        "True BusinessID Multi Domain",
        "True BusinessID with EV Multi Domain",
        "Secure Site",
        "Secure Site Pro",
        "Secure Site with EV",
        "Secure Site Pro with EV",
        "EssentialSSL",
        "EssentialSSL Wildcard",
        "InstantSSL Pro",
        "PremiumSSL Wildcard",
        "EV SSL",
        "EV SSL SGC",
        "SSL123",
        "SSL Web Server",
        "SGC Supercert",
        "SSL Webserver EV",
        "EV Multi Domain SSL",
        "Multi Domain SSL",
        "PositiveSSL Multi Domain",
        "Unified Communications",
    )

    if certificate_type not in valid_certs:
        log.error("Invalid option for certificate_type=%s", certificate_type)
        raise Exception("Invalid option for certificate_type=" + certificate_type)

    if years < 1 or years > 5:
        log.error("Invalid option for years=%s", str(years))
        raise Exception("Invalid option for years=" + str(years))

    opts = salt.utils.namecheap.get_opts("namecheap.ssl.create")

    opts["Years"] = years
    opts["Type"] = certificate_type
    if promotion_code is not None:
        opts["PromotionCode"] = promotion_code
    if sans_to_add is not None:
        opts["SANStoADD"] = sans_to_add

    response_xml = salt.utils.namecheap.post_request(opts)
    if response_xml is None:
        return {}

    sslcreateresult = response_xml.getElementsByTagName("SSLCreateResult")[0]
    sslcertinfo = sslcreateresult.getElementsByTagName("SSLCertificate")[0]

    result = salt.utils.namecheap.atts_to_dict(sslcreateresult)
    result.update(salt.utils.namecheap.atts_to_dict(sslcertinfo))
    return result


def parse_csr(csr_file, certificate_type, http_dc_validation=False):
    """
    Parses the CSR. Returns a dictionary of result values.

    csr_file
        Path to Certificate Signing Request file

    certificate_type
        Type of SSL Certificate. Possible values include:

        - EV Multi Domain SSL
        - EV SSL
        - EV SSL SGC
        - EssentialSSL
        - EssentialSSL Wildcard
        - InstantSSL
        - InstantSSL Pro
        - Multi Domain SSL
        - PositiveSSL
        - PositiveSSL Multi Domain
        - PositiveSSL Wildcard
        - PremiumSSL
        - PremiumSSL Wildcard
        - QuickSSL Premium
        - RapidSSL
        - RapidSSL Wildcard
        - SGC Supercert
        - SSL Web Server
        - SSL Webserver EV
        - SSL123
        - Secure Site
        - Secure Site Pro
        - Secure Site Pro with EV
        - Secure Site with EV
        - True BusinessID
        - True BusinessID Multi Domain
        - True BusinessID Wildcard
        - True BusinessID with EV
        - True BusinessID with EV Multi Domain
        - Unified Communications

    http_dc_validation : False
        Set to ``True`` if a Comodo certificate and validation should be
        done with files instead of emails and to return the info to do so

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' namecheap_ssl.parse_csr my-csr-file PremiumSSL
    """
    valid_certs = (
        "QuickSSL Premium",
        "RapidSSL",
        "RapidSSL Wildcard",
        "PremiumSSL",
        "InstantSSL",
        "PositiveSSL",
        "PositiveSSL Wildcard",
        "True BusinessID with EV",
        "True BusinessID",
        "True BusinessID Wildcard",
        "True BusinessID Multi Domain",
        "True BusinessID with EV Multi Domain",
        "Secure Site",
        "Secure Site Pro",
        "Secure Site with EV",
        "Secure Site Pro with EV",
        "EssentialSSL",
        "EssentialSSL Wildcard",
        "InstantSSL Pro",
        "PremiumSSL Wildcard",
        "EV SSL",
        "EV SSL SGC",
        "SSL123",
        "SSL Web Server",
        "SGC Supercert",
        "SSL Webserver EV",
        "EV Multi Domain SSL",
        "Multi Domain SSL",
        "PositiveSSL Multi Domain",
        "Unified Communications",
    )

    if certificate_type not in valid_certs:
        log.error("Invalid option for certificate_type=%s", certificate_type)
        raise Exception("Invalid option for certificate_type=" + certificate_type)

    opts = salt.utils.namecheap.get_opts("namecheap.ssl.parseCSR")

    with salt.utils.files.fopen(csr_file, "rb") as csr_handle:
        opts["csr"] = salt.utils.stringutils.to_unicode(csr_handle.read())

    opts["CertificateType"] = certificate_type
    if http_dc_validation:
        opts["HTTPDCValidation"] = "true"

    response_xml = salt.utils.namecheap.post_request(opts)

    sslparseresult = response_xml.getElementsByTagName("SSLParseCSRResult")[0]

    return salt.utils.namecheap.xml_to_dict(sslparseresult)


def get_list(**kwargs):
    """
    Returns a list of SSL certificates for a particular user

    ListType : All
        Possible values:

        - All
        - Processing
        - EmailSent
        - TechnicalProblem
        - InProgress
        - Completed
        - Deactivated
        - Active
        - Cancelled
        - NewPurchase
        - NewRenewal

        SearchTerm
            Keyword to look for on the SSL list

        Page : 1
            Page number to return

        PageSize : 20
            Total number of SSL certificates to display per page (minimum:
            ``10``, maximum: ``100``)

        SoryBy
            One of ``PURCHASEDATE``, ``PURCHASEDATE_DESC``, ``SSLTYPE``,
            ``SSLTYPE_DESC``, ``EXPIREDATETIME``, ``EXPIREDATETIME_DESC``,
            ``Host_Name``, or ``Host_Name_DESC``

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' namecheap_ssl.get_list Processing
    """
    opts = salt.utils.namecheap.get_opts("namecheap.ssl.getList")
    for key, value in kwargs.items():
        opts[key] = value

    response_xml = salt.utils.namecheap.get_request(opts)

    if response_xml is None:
        return []

    ssllistresult = response_xml.getElementsByTagName("SSLListResult")[0]

    result = []
    for e in ssllistresult.getElementsByTagName("SSL"):
        ssl = salt.utils.namecheap.atts_to_dict(e)
        result.append(ssl)

    return result


def get_info(certificate_id, returncertificate=False, returntype=None):
    """
    Retrieves information about the requested SSL certificate. Returns a
    dictionary of information about the SSL certificate with two keys:

    - **ssl** - Contains the metadata information
    - **certificate** - Contains the details for the certificate such as the
      CSR, Approver, and certificate data

    certificate_id
        Unique ID of the SSL certificate

    returncertificate : False
        Set to ``True`` to ask for the certificate in response

    returntype
        Optional type for the returned certificate. Can be either "Individual"
        (for X.509 format) or "PKCS7"

        .. note::
            Required if ``returncertificate`` is ``True``

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' namecheap_ssl.get_info my-cert-id
    """
    opts = salt.utils.namecheap.get_opts("namecheap.ssl.getinfo")
    opts["certificateID"] = certificate_id

    if returncertificate:
        opts["returncertificate"] = "true"
        if returntype is None:
            log.error(
                "returntype must be specified when returncertificate is set to True"
            )
            raise Exception(
                "returntype must be specified when returncertificate is set to True"
            )
        if returntype not in ["Individual", "PKCS7"]:
            log.error(
                "returntype must be specified as Individual or PKCS7, not %s",
                returntype,
            )
            raise Exception(
                "returntype must be specified as Individual or PKCS7, not " + returntype
            )
        opts["returntype"] = returntype

    response_xml = salt.utils.namecheap.get_request(opts)

    if response_xml is None:
        return {}

    sslinforesult = response_xml.getElementsByTagName("SSLGetInfoResult")[0]

    return salt.utils.namecheap.xml_to_dict(sslinforesult)



"""
NAPALM Grains
=============

:codeauthor: Mircea Ulinic <ping@mirceaulinic.net>
:maturity:   new
:depends:    napalm
:platform:   unix

Dependencies
------------

- :mod:`NAPALM proxy module <salt.proxies.napalm>`

.. versionadded:: 2016.11.0
"""


import logging

import salt.utils.dns
import salt.utils.napalm

log = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------------------------------------
# grains properties
# ----------------------------------------------------------------------------------------------------------------------

__virtualname__ = "napalm"
__proxyenabled__ = ["napalm"]

# ----------------------------------------------------------------------------------------------------------------------
# global variables
# ----------------------------------------------------------------------------------------------------------------------

GRAINS_CACHE = {}
DEVICE_CACHE = {}

_FORBIDDEN_OPT_ARGS = [
    "secret",  # used by IOS to enter in enable mode
    "enable_password",  # used by EOS
]

# ----------------------------------------------------------------------------------------------------------------------
# property functions
# ----------------------------------------------------------------------------------------------------------------------


def __virtual__():
    """
    NAPALM library must be installed for this module to work and run in a (proxy) minion.
    """
    return salt.utils.napalm.virtual(__opts__, __virtualname__, __file__)


# ----------------------------------------------------------------------------------------------------------------------
# helpers
# ----------------------------------------------------------------------------------------------------------------------


def _retrieve_grains_cache(proxy=None):
    """
    Retrieves the grains from the network device if not cached already.
    """
    global GRAINS_CACHE
    if not GRAINS_CACHE:
        if proxy and salt.utils.napalm.is_proxy(__opts__):
            # if proxy var passed and is NAPALM-type proxy minion
            GRAINS_CACHE = proxy["napalm.get_grains"]()
        elif not proxy and salt.utils.napalm.is_minion(__opts__):
            # if proxy var not passed and is running in a straight minion
            GRAINS_CACHE = salt.utils.napalm.call(DEVICE_CACHE, "get_facts", **{})
    return GRAINS_CACHE


def _retrieve_device_cache(proxy=None):
    """
    Loads the network device details if not cached already.
    """
    global DEVICE_CACHE
    if not DEVICE_CACHE:
        if proxy and salt.utils.napalm.is_proxy(__opts__):
            # if proxy var passed and is NAPALM-type proxy minion
            if "napalm.get_device" in proxy:
                DEVICE_CACHE = proxy["napalm.get_device"]()
        elif not proxy and salt.utils.napalm.is_minion(__opts__):
            # if proxy var not passed and is running in a straight minion
            DEVICE_CACHE = salt.utils.napalm.get_device(__opts__)
    return DEVICE_CACHE


def _get_grain(name, proxy=None):
    """
    Retrieves the grain value from the cached dictionary.
    """
    grains = _retrieve_grains_cache(proxy=proxy)
    if grains.get("result", False) and grains.get("out", {}):
        return grains.get("out").get(name)


def _get_device_grain(name, proxy=None):
    """
    Retrieves device-specific grains.
    """
    device = _retrieve_device_cache(proxy=proxy)
    return device.get(name.upper())


# ----------------------------------------------------------------------------------------------------------------------
# actual grains
# ----------------------------------------------------------------------------------------------------------------------


def getos(proxy=None):
    """
    Returns the Operating System name running on the network device.

    Example: junos, iosxr, eos, ios etc.

    CLI Example - select all network devices running JunOS:

    .. code-block:: bash

        salt -G 'os:junos' test.ping
    """
    return {"os": _get_device_grain("driver_name", proxy=proxy)}


def version(proxy=None):
    """
    Returns the OS version.

    Example: 13.3R6.5, 6.0.2 etc.

    CLI Example - select all network devices running JunOS 13.3R6.5 and return the model:

    .. code-block:: bash

        salt -G 'os:junos and version:13.3R6.5' grains.get model

    Output:

    .. code-block:: yaml

        edge01.bjm01:
            MX2000
        edge01.sjc01:
            MX960
        edge01.mrs01:
            MX480
        edge01.muc01:
            MX240
    """
    return {"version": _get_grain("os_version", proxy=proxy)}


def model(proxy=None):
    """
    Returns the network device chassis model.

    Example: MX480, ASR-9904-AC etc.

    CLI Example - select all Juniper MX480 routers and execute traceroute to 8.8.8.8:

    .. code-block:: bash

        salt -G 'model:MX480' net.traceroute 8.8.8.8
    """
    return {"model": _get_grain("model", proxy=proxy)}


def serial(proxy=None):
    """
    Returns the chassis serial number.

    Example: FOX1234W00F

    CLI Example - select all devices whose serial number begins with `FOX` and display the serial number value:

    .. code-block:: bash

        salt -G 'serial:FOX*' grains.get serial

    Output:

    .. code-block:: yaml

        edge01.icn01:
            FOXW00F001
        edge01.del01:
            FOXW00F002
        edge01.yyz01:
            FOXW00F003
        edge01.mrs01:
            FOXW00F004
    """
    return {"serial": _get_grain("serial_number", proxy=proxy)}


def vendor(proxy=None):
    """
    Returns the network device vendor.

    Example: juniper, cisco, arista etc.

    CLI Example - select all devices produced by Cisco and shutdown:

    .. code-block:: bash

        salt -G 'vendor:cisco' net.cli "shut"
    """
    return {"vendor": _get_grain("vendor", proxy=proxy)}


def uptime(proxy=None):
    """
    Returns the uptime in seconds.

    CLI Example - select all devices started/restarted within the last hour:

    .. code-block:: bash

        salt -G 'uptime<3600' test.ping
    """
    return {"uptime": _get_grain("uptime", proxy=proxy)}


def interfaces(proxy=None):
    """
    Returns the complete interfaces list of the network device.

    Example: ['lc-0/0/0', 'pfe-0/0/0', 'xe-1/3/0', 'lo0', 'irb', 'demux0', 'fxp0']

    CLI Example - select all devices that have a certain interface, e.g.: xe-1/1/1:

    .. code-block:: bash

        salt -G 'interfaces:xe-1/1/1' test.ping

    Output:

    .. code-block:: yaml

        edge01.yyz01:
            True
        edge01.maa01:
            True
        edge01.syd01:
            True
        edge01.del01:
            True
        edge01.dus01:
            True
        edge01.kix01:
            True
    """
    return {"interfaces": _get_grain("interface_list", proxy=proxy)}


def username(proxy=None):
    """
    Return the username.

    .. versionadded:: 2017.7.0

    CLI Example - select all devices using `foobar` as username for connection:

    .. code-block:: bash

        salt -G 'username:foobar' test.ping

    Output:

    .. code-block:: yaml

        device1:
            True
        device2:
            True
    """
    if proxy and salt.utils.napalm.is_proxy(__opts__):
        # only if proxy will override the username
        # otherwise will use the default Salt grains
        return {"username": _get_device_grain("username", proxy=proxy)}


def hostname(proxy=None):
    """
    Return the hostname as configured on the network device.

    CLI Example:

    .. code-block:: bash

        salt 'device*' grains.get hostname

    Output:

    .. code-block:: yaml

        device1:
            edge01.yyz01
        device2:
            edge01.bjm01
        device3:
            edge01.flw01
    """
    return {"hostname": _get_grain("hostname", proxy=proxy)}


def host(proxy=None):
    """
    This grain is set by the NAPALM grain module
    only when running in a proxy minion.
    When Salt is installed directly on the network device,
    thus running a regular minion, the ``host`` grain
    provides the physical hostname of the network device,
    as it would be on an ordinary minion server.
    When running in a proxy minion, ``host`` points to the
    value configured in the pillar: :mod:`NAPALM proxy module <salt.proxy.napalm>`.

    .. note::

        The diference between ``host`` and ``hostname`` is that
        ``host`` provides the physical location - either domain name or IP address,
        while ``hostname`` provides the hostname as configured on the device.
        They are not necessarily the same.

    .. versionadded:: 2017.7.0

    CLI Example:

    .. code-block:: bash

        salt 'device*' grains.get host

    Output:

    .. code-block:: yaml

        device1:
            ip-172-31-13-136.us-east-2.compute.internal
        device2:
            ip-172-31-11-193.us-east-2.compute.internal
        device3:
            ip-172-31-2-181.us-east-2.compute.internal
    """
    if proxy and salt.utils.napalm.is_proxy(__opts__):
        # this grain is set only when running in a proxy minion
        # otherwise will use the default Salt grains
        return {"host": _get_device_grain("hostname", proxy=proxy)}


def host_dns(proxy=None):
    """
    Return the DNS information of the host.
    This grain is a dictionary having two keys:

    - ``A``
    - ``AAAA``

    .. note::
        This grain is disabled by default, as the proxy startup may be slower
        when the lookup fails.
        The user can enable it using the ``napalm_host_dns_grain`` option (in
        the pillar or proxy configuration file):

        .. code-block:: yaml

            napalm_host_dns_grain: true

    .. versionadded:: 2017.7.0

    CLI Example:

    .. code-block:: bash

        salt 'device*' grains.get host_dns

    Output:

    .. code-block:: yaml

        device1:
            A:
                - 172.31.9.153
            AAAA:
                - fd52:188c:c068::1
        device2:
            A:
                - 172.31.46.249
            AAAA:
                - fdca:3b17:31ab::17
        device3:
            A:
                - 172.31.8.167
            AAAA:
                - fd0f:9fd6:5fab::1
    """
    if not __opts__.get("napalm_host_dns_grain", False):
        return
    device_host = host(proxy=proxy)
    if device_host:
        device_host_value = device_host["host"]
        host_dns_ret = {"host_dns": {"A": [], "AAAA": []}}
        dns_a = salt.utils.dns.lookup(device_host_value, "A")
        if dns_a:
            host_dns_ret["host_dns"]["A"] = dns_a
        dns_aaaa = salt.utils.dns.lookup(device_host_value, "AAAA")
        if dns_aaaa:
            host_dns_ret["host_dns"]["AAAA"] = dns_aaaa
        return host_dns_ret


def optional_args(proxy=None):
    """
    Return the connection optional args.

    .. note::

        Sensible data will not be returned.

    .. versionadded:: 2017.7.0

    CLI Example - select all devices connecting via port 1234:

    .. code-block:: bash

        salt -G 'optional_args:port:1234' test.ping

    Output:

    .. code-block:: yaml

        device1:
            True
        device2:
            True
    """
    opt_args = _get_device_grain("optional_args", proxy=proxy) or {}
    if opt_args and _FORBIDDEN_OPT_ARGS:
        for arg in _FORBIDDEN_OPT_ARGS:
            opt_args.pop(arg, None)
    return {"optional_args": opt_args}
